
    // Functions for handling strings
    // MathOnNapkins Copyright 2010
    
    #include <stdio.h>
    #include <string.h>

    #include <exception>
    
    #include "buffer.h"

    using std::exception;

// ===============================================================

    // Static (class-wide) variables of Buffer
    int Buffer::allocations = 0;
    int Buffer::totalMem    = 0;
    
// ===============================================================

    Buffer::Buffer(int length)
    {
        InitBuffer(length);
    }

// ===============================================================

    Buffer::Buffer(int width, int height, int unit)
    {
        InitBuffer(width * height * unit);
        
        this->width  = width;
        this->height = height;
        this->unit   = unit;
    }

// ===============================================================

    void Buffer::InitBuffer(int length)
    {
        this->error    = 0;
        
        this->length   = length;
        
        this->height   = 1;
        this->width    = length;
        this->unit     = 1;
        
        this->rOffset  = 0;
        this->wOffset  = 0;
        
        if(length <= 0)
        {
            throw exception();;
        }
        
        this->contents = new unsigned char[length + 1];
        
        allocations += 1;
        
        memset(this->contents, 0, length);
        
        totalMem += (length + 1);
    }

// ===============================================================

    Buffer* Buffer::Duplicate()
    {
        Buffer *b = NULL;

        // -----------------------

        b = new Buffer(this->width, this->height, this->unit);
        
        memcpy(b->contents, this->contents, this->length);
        
        return b;
    }

// ===============================================================

    void Buffer::Resize(uint32_t newSize)
    {
        unsigned char *newContents;
        
        // ------------------------------
    
        if(newSize <= 0)
        {
            throw exception();
            
            return;
        }
        
        newContents = new unsigned char[newSize + 1];
        
        memset(newContents, 0, newSize + 1);
        
        if(newSize > this->length)
        {
            memcpy(newContents, this->contents, this->length);
        }
        else
        {
            memcpy(newContents, this->contents, newSize);
        }
        
        totalMem -= this->length;
        totalMem += newSize;

        delete this->contents;
        
        this->contents = newContents;
        this->length   = newSize;
    }

// ===============================================================

    Buffer::~Buffer()
    {
        allocations -= 1;
        totalMem    -= this->length + 1;
        
        delete this->contents;
    }

// ===============================================================

    void Buffer::Concat(bufPtr first, bufPtr second)
    {
    #if(0)
    
        // needs work but too lazy.
        int newBufSize = (first->length + second->length + 1) * charSize;
        int counter = first->length;
        
        // temporary buffer used to extend the 
        unsigned char* temp = (unsigned char*) malloc(newBufSize);
        
        if(!temp)
        {
            MessageBox(0, "ConcatBuffer Failure", "error", MB_OK);
            return 0;
        }
        
        // Only copy the relevant bytes. Ignore the zero terminator.
        memcpy(temp, first->contents, first->length);
        
        // Copy the second buffer where the first one ends, then add in the null terminator.
        for(counter ; counter < newBufSize ; counter += charSize)
                temp[counter] = second->contents[counter - (first->length)];
        
        
        // might generate an error. not sure.
        free(second->contents);
        
        //Also might generate an error.
        free(first->contents);
        
        first->contents = temp;
        first->length = newBufSize;
        
        return 1;
        
    #endif
    }

// ===============================================================

    uint32_t Buffer::GetByte()
    {
        this->rOffset += 1;
        
        return GetByte(this->rOffset - 1);
    }

// ===============================================================

    uint32_t Buffer::GetByte(int offset)
    {   
        // returns an 8 bit integer
        int checker = offset + 1;
        
        // ----------------------------
        
        if(offset < 0)
        {
            throw exception();
            
            return 0;
        }
        else if( (unsigned int) checker > this->length)
        {
            throw exception();
            
            return 0;
        }
        
        checker = this->contents[offset];
        
        checker &= 0xFF;
        
        return checker;
    }

// ===============================================================

    uint32_t Buffer::GetByte(uint32_t x, uint32_t y)
    {
        return GetByte( (x + (y * this->width)) * this->unit );
    }
    
// ===============================================================

    uint32_t Buffer::Get2Bytes(int offset)
    {
        // returns a 16 bit integer
        
        int checker = offset + 2;
        
        // ----------------------
        
        if(offset < 0)
        {
            throw exception();
            
            return 0;
        }
        else if( (unsigned int) checker > this->length)
        {
            throw exception();
            
            return 0;
        }
        
        checker  = ( (this->contents[offset + 0] <<  0) & 0x000000FF );
        checker |= ( (this->contents[offset + 1] <<  8) & 0x0000FF00 );
        
        checker &= 0xFFFF;
        
        return checker;
    }

// ===============================================================

    uint32_t Buffer::Get2Bytes(uint32_t x, uint32_t y)
    {
        return Get2Bytes( (x + (y * this->width)) * this->unit );
    }

// ===============================================================

    uint32_t Buffer::Get3Bytes(int offset)
    {
        // returns a 24-bit integer
        
        int checker = offset + 3;
        
        // -------------------------------
        
        if(offset < 0)
        {
            throw exception();
            
            return 0;
        }
        else if( (unsigned int) checker > this->length)
        {
            throw exception();
            
            return 0;
        }
        
        checker  = ( (this->contents[offset + 0] <<  0) & 0x000000FF );
        checker |= ( (this->contents[offset + 1] <<  8) & 0x0000FF00 );
        checker |= ( (this->contents[offset + 2] << 16) & 0x00FF0000 );
        
        checker &= 0xFFFFFF;
        
        return checker;
    }

// ===============================================================

    uint32_t Buffer::Get3Bytes(uint32_t x, uint32_t y)
    {
        return Get3Bytes( (x + (y * this->width)) * this->unit );
    }

// ===============================================================

    uint32_t Buffer::Get4Bytes()
    {
        uint32_t value   = 0;
        
        // ------------------------
        
        value  |= (GetByte() <<  0);
        value  |= (GetByte() <<  8);
        value  |= (GetByte() << 16);
        value  |= (GetByte() << 24);
        
        return value;
    }

// ===============================================================

    uint32_t Buffer::Get4Bytes(uint32_t offset)
    {
        uint32_t checker = offset + 4;
        
        // ----------------------------
        
        if(offset < 0)
        {
            throw exception();
            
            return 0;
        }
        else if( checker > this->length)
        {
            throw exception();
            
            return 0;
        }
        
        checker  = ( (this->contents[offset + 0] <<  0) & 0x000000FF );
        checker |= ( (this->contents[offset + 1] <<  8) & 0x0000FF00 );
        checker |= ( (this->contents[offset + 2] << 16) & 0x00FF0000 );
        checker |= ( (this->contents[offset + 3] << 24) & 0xFF000000 );
        
        return checker;
    }
    
// ===============================================================

    uint32_t Buffer::Get4Bytes(uint32_t x, uint32_t y)
    {
        return Get4Bytes( (x + (y * this->width)) * this->unit );
    }

// ===============================================================

    uint32_t Buffer::GetBit(uint32_t bit)
    {
        uint8_t mask = 0;
        
        uint32_t offset    = (bit / 8);
        uint32_t remainder = (bit % 8);
        uint32_t checker   = offset + 1;
        
        // ----------------------------

        if( checker > this->length)
        {
            throw exception();
            
            return 0;
        }
        
        mask  = this->contents[offset];
        mask &= (1 << remainder);
        
        return (mask ? 1 : 0);
    }

// ===============================================================

    void Buffer::PutBit(uint32_t value, uint32_t bit)
    {
        // checker is the byte of the buffer to look at.
        // remainder is 0 to 7 and helps us find the

        uint8_t mask = 0;
        
        uint32_t offset    = (bit / 8);
        uint32_t remainder = (bit % 8); 
        uint32_t checker   = offset + 1;
        
        // --------------------------
        
        if( checker > this->length)
        {
            throw exception();
        }
        
        mask = 1 << remainder;
        this->contents[offset] &= ~mask;

        mask = value ? mask : 0;
        this->contents[offset] |= mask;
    }

// ===============================================================

    void Buffer::PutByte(uint32_t offset, uint32_t value)
    {
        uint32_t checker = offset + 1;
        
        // -------------------------------
        
        if(offset < 0)
        {
            throw exception();
        }
        else if( checker > this->length)
        {
            throw exception();
        }
        
        value &= 0xFF;
        
        this->contents[offset] = (uint8_t) value;
    }

// ===============================================================

    void Buffer::PutByte(int value)
    {
        this->wOffset += 1;
    
        PutByte(this->wOffset - 1, value);
    }

// ===============================================================

    void Buffer::PutByte(uint32_t x, uint32_t y, uint32_t value)
    {
        PutByte( (x + (y * this->width)) * this->unit, value);
    }

// ===============================================================

    void Buffer::Put2Bytes(uint32_t *offset, uint32_t value)
    {    
        if(offset)
        {
            Put2Bytes( (*offset), value );

            (*offset) += 2;
        }
    }

// ===============================================================

    void Buffer::Put2Bytes(uint32_t value)
    {
        Put2Bytes(this->wOffset, value);
        
        this->wOffset += 2;
    }

// ===============================================================

    void Buffer::Put2Bytes(uint32_t offset, uint32_t value)
    {
        uint32_t checker = offset + 2;
        
        // -----------------------------------

        if( checker > this->length)
        {
            throw exception();
        }
        
        this->contents[offset + 0] = (char) ((value & 0x000000FF) >> 0);
        this->contents[offset + 1] = (char) ((value & 0x0000FF00) >> 8);        
    }

// ===============================================================

    void Buffer::Put2Bytes(uint32_t x, uint32_t y, uint32_t value)
    {
        Put2Bytes( (x + (y * this->width)) * this->unit, value);
    }

// ===============================================================

    void Buffer::Put3Bytes(int offset, int value)
    {
        uint32_t checker = offset + 3;
        
        // ----------------------------
        
        if(offset < 0)
        {
            throw exception();
        }
        else if( checker > this->length)
        {
            throw exception();
        }
        
        this->contents[offset + 0] = (char) ((value & 0x000000FF) >>  0);
        this->contents[offset + 1] = (char) ((value & 0x0000FF00) >>  8);        
        this->contents[offset + 2] = (char) ((value & 0x00FF0000) >> 16);        
    }

// ===============================================================

    void Buffer::Put3Bytes(uint32_t x, uint32_t y, uint32_t value)
    {
        Put3Bytes( (x + (y * this->width)) * this->unit, value);
    }

// ===============================================================

    void Buffer::Put4Bytes(int offset, int value)
    {
        uint32_t checker = offset + 4;
        
        // ------------------------------------
        
        if(offset < 0)
        {
            throw exception();;
        }
        else if(checker > this->length)
        {
            throw exception();;
        }
        
        this->contents[offset + 0] = (char) ((value & 0x000000FF) >>  0);
        this->contents[offset + 1] = (char) ((value & 0x0000FF00) >>  8);
        this->contents[offset + 2] = (char) ((value & 0x00FF0000) >> 16);
        this->contents[offset + 3] = (char) ((value & 0xFF000000) >> 24);
    }

// ===============================================================

    void Buffer::Put4Bytes(uint32_t value)
    {
        int byte1   = (value & 0x000000FF)  >> 0x00;
        int byte2   = (value & 0x0000FF00)  >> 0x08;
        int byte3   = (value & 0x00FF0000)  >> 0x10;
        int byte4   = (value & 0xFF000000)  >> 0x18;
        
        // ----------------------------------------
        
        // a failure of any of these will cause result to fall to zero.
        PutByte(byte1);
        PutByte(byte2);
        PutByte(byte3);
        PutByte(byte4);
    }

// ===============================================================

    void Buffer::Put4Bytes(uint32_t x, uint32_t y, uint32_t value)
    {
        Put4Bytes( (x + (y * this->width)) * this->unit, value );
    }


// ===============================================================

    void Buffer::ToFile(char* pathName, uint32_t length, uint32_t offset)
    {
        FILE* f = NULL;
        
        // --------------------------
        
        f = fopen(pathName, "wb");
        
        if(!f)
        {
            throw exception();
        }
        
        if(length)
        {
            // set to safe defaults if the length and offset values don't make sense
            if( (length + offset) > this->length)
            {
                length = this->length;
                offset = 0;
            }            
            
            fwrite(this->contents + offset, length, 1, f);
        }
        else
        {   
            // If length is set to zero, by default we just write the whole buffer to a file.
            fwrite(this->contents, this->length, 1, f);
        }
        
        fclose(f);
    }

// ===============================================================

    void Buffer::FromFile(char* pathName, uint32_t length, uint32_t offset)
    {
        uint32_t fileLength = 0;
        
        FILE* f = NULL;
        
        // --------------------------
        
        f = fopen(pathName, "rb");
        
        if( (this->length == 0) || !f)
        {
            throw exception();
            
            return;
        }
        
        fseek(f, 0, SEEK_END);
        
        fileLength  = ftell(f);
        
        fseek(f, 0, SEEK_SET);
        
        if(fileLength > this->length)
        {
            throw exception();
            
            return;
        }
        
        fread(this->contents + offset, fileLength, 1, f);
        
        fclose(f);
    }
   
// ===============================================================

    void Buffer::Copy(Buffer *target, Buffer *src, unsigned int tOffset, unsigned int sOffset, unsigned int numBytes)
    {
        if( !target || !src)
        {
            throw exception();
            
            return;
        }
        else if( !target->contents || !src->contents )
        {
            throw exception();
            
            return;
        }
        else if( (tOffset + numBytes) > target->length )
        {
            throw exception();
            
            return;
        }
        else if( (sOffset + numBytes) > src->length )
        {
            throw exception();
            
            return;
        }
        
        memcpy(target->contents + tOffset, src->contents + sOffset, numBytes);
    } 

// ===============================================================
    
    // An auto-incrementing version of CopyBuffer()
    
    void Buffer::Copy(Buffer *target, Buffer *src, uint32_t numBytes)
    {
        if( !target || !src)
        {
            throw exception();
            
            return;
        }
        
        Copy(target, src, target->wOffset, 0, numBytes);
        
        target->wOffset += numBytes;
    }

// ===============================================================

    // block copy version of CopyBuffer()
    
    void Buffer::Copy(bufPtr target,  bufPtr src,
                      uint32_t xTarg, uint32_t yTarg,
                      uint32_t xSrc,  uint32_t ySrc,
                      uint32_t width, uint32_t height)
    {
        uint32_t i = 0, j = 0, result = 0;
        
        // -----------------------------------------

        if( !target || !src)
        {
            throw exception();
            
            return;
        }
        else if(target->unit != src->unit)
        {
            throw exception();
            
            return;
        }

        for(i = 0; i < width; ++i)
        {
            for(j = 0; j < height; ++j)
            {
                Copy(target,
                     src, 
                     (xTarg + (yTarg * target->width)) * target->unit,
                     (xSrc + (ySrc * src->width)) * src->unit,
                     src->unit);
            }
        }
    }

// ===============================================================

    // The WB stands for "write back" meaning it will update the pointers automatically
    // After writing them
    
    void Buffer::CopyWB(bufPtr target, bufPtr src, uint32_t *tPtr, uint32_t *sPtr, uint32_t numBytes)
    {
        Copy(target, src, *tPtr, *sPtr, numBytes);
        
        *tPtr += numBytes;
        *sPtr += numBytes;
    }    

// ===============================================================

    Buffer* Buffer::GetSubBuffer(uint32_t offset, uint32_t newLength)
    {
        uint32_t finalLength = 0;
        
        Buffer* subBuffer    = NULL;
        
        // ----------------------------------
        
        if(offset > this->length - 1)
        {
            throw exception();
            
            return NULL;
        }
        
        if(offset + newLength > this->length)
        {
            throw exception();
            
            return NULL;
        }
        
        if(newLength)
        {
            finalLength = newLength;
        }
        else
        {
            finalLength = this->length - offset;
        }
        
        subBuffer = new Buffer(finalLength);
        
        subBuffer->unit = this->unit;
        
        memcpy(subBuffer->contents, &(this->contents[offset]), finalLength);
        
        /// todo: if buffer is a grid, try to mimic its grid structure in the subbuffer?
        
        return subBuffer; 
    }
    
// ===============================================================

#if(0)


// ===============================================================

    void Buffer::ZeroBuffer()
    {
        memcpy(this->contents, 0, this->length);
    }

// ===============================================================

    bool Buffer::CompareBuffer(bufPtr a, bufPtr b)
    {
        // just a wrapper around memcmp really

        if(!a || !b)
        {
            throw exception();
            
            return false;
        }
        
        if(a->length != b->length)
            return false;
        
        // memcmp returns 0 if the memory is the same, so I'm returning
        // the boolean version of that. (true if they are the same)
        return memcmp(a->contents, b->contents, a->length) ? false : true; 
    }

// ===============================================================

    uint32_t hextodec(u8 *input, uint32_t length)
    {
        uint32_t result = 0;
	    uint32_t value;
        uint32_t ceiling = strlen((char*) input) - 1;
	    uint32_t power16 = 0;

        // ceiling is intepreted as the highest number of digits - 1 to handle
        // j is a counter through the array that starts at the end
        // (from our perspective us the lowest place)
        // length is the largest number of digits we want back
        // since the return value is an int, we can't have
        // length be any longer than 8
        // but with that leaves the question of... are there higher
        // digits that would add to the value and cause it to overflow
        // if that is the case I'm going to return 0xFFFFFFFF

        length = length > 8 ? 8 : length - 1;

        length = ceiling < length ? ceiling : length;

	    int j = ceiling;    

	    for( ;  j >= 0; j--)
	    {       
		    if(input[j] >= 'A' && input[j] <= 'F')
		    {
			    value = input[j] - 'F';
			    value += 15;
		    }
   		    else if(input[j] >= 'a' && input[j] <= 'f')
		    {
			    value = input[j] - 'f';
			    value += 15;
		    }
		    else if(input[j] >= '0' && input[j] <= '9')
		    {
			    value = input[j] - '9';
			    value += 9;
		    }
            else
                value = 0;

            if(j >= (ceiling - length))
            {
                result += (value << (4 * power16));
		        power16++;

                continue;
            }
            
            if(length != 7)
                break;
            
            // at this point the useful data has expired
            // we're just looking for overflow 
            if(value)
            {
                result = 0xFFFFFFFF;
                break;
            }


	    }
	    
	    return result;
    }

// ===============================================================

    void makeHexStr(bufPtr input, uint32_t value)
    {
        char temp[8];

        uint32_t a = min(sizeof(temp), input->length);

        value &= 0xFFFFFFFF;

        itoa(value, temp, 0x10);
        strupr(temp);

        strncpy((char*) input->contents, temp, a);
    }

// ===============================================================

    unsigned char* ToString(bufPtr input)
    {
        return input->contents;
    }

// ===============================================================

    bufPtr FromString(char *contents, int length)
    {
        bufPtr newBuf = CreateBuffer(length);

        if(newBuf)
            memcpy(newBuf->contents, contents, length);

        return newBuf;
    }

// ===============================================================

    bufPtr FromString(char *contents, int width, int height, int unit)
    {
        bufPtr newBuf =  CreateBuffer(width, height, unit);

        if(newBuf)
            memcpy(newBuf->contents, contents, (width * height * unit) );
       
        return newBuf;
    }


// ===============================================================

    int GetHexEditNum(HWND win)
    {
        char dummy[256];
    
        GetWindowText(win, dummy, 256);
    
        return hextodec((u8*) dummy, strlen(dummy));
    }

// ===============================================================

#endif
