    
    #include "inttypes.h"
    
#ifndef BUFFER_H

    #define BUFFER_H

    // Variables
    
    // structure types
    class Buffer 
    {
    public:
    
        static int allocations;
        static int totalMem;
        
        unsigned char* contents;
        unsigned int length;
        unsigned int error;
      
        // write offset
      
        unsigned int wOffset;
        
        // read offset        
        unsigned int rOffset;
        
        // "unit" is added in case the internal data is in X byte chunks
        unsigned int width;
        unsigned int height;
        unsigned int unit;
    
    public:
        
        Buffer(int length);
        Buffer(int width, int height, int uint);
        ~Buffer();
        
        void    InitBuffer(int length);
        
        Buffer* Duplicate();

        uint32_t Get4Bytes();
        uint32_t Get4Bytes(uint32_t offset);
        uint32_t Get4Bytes(uint32_t x, uint32_t y);

        uint32_t Get3Bytes(int offset);
        uint32_t Get3Bytes(uint32_t x, uint32_t y);

        uint32_t Get2Bytes(int offset);
        uint32_t Get2Bytes(uint32_t x, uint32_t y);

        uint32_t GetByte();
        uint32_t GetByte(int offset);
        uint32_t GetByte(uint32_t x, uint32_t y);
        
        void    FromFile(char* pathName, uint32_t length = 0, uint32_t offset = 0);
        void    ToFile(char* pathName,   uint32_t length = 0, uint32_t offset = 0);

        uint32_t GetBit(uint32_t bit);
        
        void  PutBit(uint32_t value, uint32_t bit);

        void    PutByte(int value);
        void    PutByte(uint32_t offset, uint32_t value);
        void    PutByte(uint32_t x, uint32_t y, uint32_t value);

        void    Put2Bytes(uint32_t *offset, uint32_t value);
        void    Put2Bytes(uint32_t value);
        void    Put2Bytes(uint32_t offset, uint32_t value);
        void    Put2Bytes(uint32_t x, uint32_t y, uint32_t value);

        void    Put3Bytes(int offset, int value);
        void    Put3Bytes(uint32_t x, uint32_t y, uint32_t value);

        void    Put4Bytes(int offset, int value);
        void    Put4Bytes(uint32_t value);
        void    Put4Bytes(uint32_t x, uint32_t y, uint32_t value);

        void    Resize(uint32_t newSize);        
        
        Buffer* GetSubBuffer(uint32_t offset, uint32_t newLength = 0);
        
         // static functions
        static void Concat(Buffer *first, Buffer* second);
        static void Copy(Buffer *target, Buffer *src, uint32_t numBytes);
        static void Copy(Buffer *target, Buffer *src, uint32_t tOffset, uint32_t sOffset, uint32_t numBytes);
        
        static void Copy(Buffer *target, Buffer *src,
                               uint32_t xTarg, uint32_t yTarg,
                               uint32_t xSrc,  uint32_t ySrc,
                               uint32_t width, uint32_t height);
        
        static void CopyWB(Buffer *target, Buffer *src, uint32_t *tPtr, uint32_t *sPtr, uint32_t numBytes);
        
        
    };

    typedef Buffer*     bufPtr;

    // function prototypes *************************************
    
#if(0)

    int     ResizeBuffer(bufPtr bufArg, int newSize);


    int     ZeroBuffer(bufPtr arg);

    bool    CompareBuffer(bufPtr a, bufPtr b);

    bufPtr  FromString(char *contents, int length);
    bufPtr  FromString(char *contents, int width, int height, int unit);

    unsigned char* ToString(bufPtr input);

    unsigned int GetBank(unsigned int cpuAddr);
    int     IsBankChange(unsigned int cpuAddr, int size);



    // Black Magic.cpp routines

    void    MainLoop();

    void    MakeNewEditWindow();

    long _stdcall MainProc(HWND, UINT, WPARAM, LPARAM);

    int IsDuplicateRom(char compareBuf[MAX_PATH]);

    int GetFile(OPENFILENAME *ofn, HWND win);

    void MakeNewGamePtr();

    int SaveGame(HWND win);

    //void LoadDataFromRom();

    void DeletePointer();

    void DrawBitmap(HWND win);

    int CALLBACK ExpandProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

    int DlgCommand(HWND dlg, WPARAM wParam, LPARAM lParam);

    void ExpandRom(int *newSize, int force);

    void OffsetWindowIds(WindowElement *we);

    void AssignCurrentWES(WindowElement *we);

    void PopulateCombobox(HWND combobox, char **stringList);

    int CALLBACK NewEditProc(HWND edit, UINT msg, WPARAM wParam, LPARAM lParam);

    void ActivateCurrentWES(WEStruct *currentWES, HWND parent);

    uint32_t hextodec(u8 *input, uint32_t length);

    void makeHexStr(bufPtr input, uint32_t value);

    int GetHexEditNum(HWND win);

    int dung(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

    long _stdcall PictureProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

    BOOL CALLBACK MoveChildren(HWND hwndChild, LPARAM lParam);
    
#endif

#endif

