
    #include "inttypes.h"

    // Constants.h 
    // contains, you guessed it, constant definitions

#ifndef CONSTANTS_H

    #define CONSTANTS_H

    extern const int charSize;

    // size of a megabit, in terms of bytes.
    extern const int _MEGABIT;
    extern const int _4MEGABITS;
    extern const int _MEGABYTE;

    // dungeon asm pointers -------------------------------------

    extern const int asm_header_ref;
    extern const int asm_header_bank;
    
    extern const int asm_layout_ref;
    extern const int asm_layout_ref2;
    
    extern const int asm_object_ref;
    extern const int asm_object_ref_2;
    extern const int asm_object_ref_3;
    extern const int asm_object_ref_4;

    extern const int asm_entrance_ref;

    extern const int asm_num_chests;
    extern const int asm_chest_ref;

    extern const int asm_sprite_ref;
    extern const int asm_sprite_hook;
    
    extern const int asm_push_block_ref;

    // #define asm_sprite_local   0x4CEE7 Shards of Might's location (doesn't work anyway)
    extern const int asm_sprite_local;

    extern const int asm_text_ref;
    extern const int asm_text_ref2;
    extern const int asm_code_lengths;

    // overworld asm pointers -------------------------------------

    // table of 24-bit pointers to the first half of each overworld map
    // In a clean rom this should be 0x2F94D (0x1794D in rom)
    extern const int asm_overmap_ptrs;

    // table of 24-bit pointers to the second half of each overworld map
    // In a clean rom this should be 0x2F5C9 (0x175C9 in rom)
    extern const int asm_overmap_ptrs2;

    // if the map16 loader is untouched these should all be 0x038000
    extern const int asm_map32To16UL;
    extern const int asm_map32To16UR;
    extern const int asm_map32To16LL;
    extern const int asm_map32To16LR;

    // defaults to $1BB96F ($DB96F in rom)
    extern const int asm_entrance_ptr_old;

    // in a clean rom this is 0xB800 (2 bytes) which references 0x1BB800 (0xDB800) in rom
    extern const int asm_hole_ptr_old;

    // graphics asm pointers ---------------------------------------

    // ROM addresses of asm location where an untouched Zelda 3 rom will load graphics from
    extern const int asm_spr_gfx_ptr;
    extern const int asm_bg_gfx_ptr;

    extern const int default_spr_gfx_ptr;
    extern const int default_bg_gfx_ptr;
    
    // black magic specific pointers ------------------------------
    extern const int expand_info_loc;
    
    // assembler macros for generating code -----------------------    
    
    #define _ASLA               0x0A
    
    #define _REP(x)             0xC2,x
    #define _SEP(x)             0xE2,x
    
    #define _LDAD(x)            0xA5,x
    #define _LDAI(x)            0xA9,x
    #define _LDALX(x,y,z)       0xBF,z,y,x
    #define _LDAX(x,y)          0xBD,y,x
    #define _LDAY(x)            0xB9,x
    #define _LDAA(x,y)          0xAD,y,x
    
    #define _STADP(x)           0x85,x
    
    #define _PEA(x,y)           0xF4,y,x
    
    #define _TAY                0xA8
    #define _TAX                0xAA
    #define _TXA                0x8A
    
    #define _JML(x,y,z)         0x5C,z,y,x
    #define _JSL(x,y,z)         0x22,z,y,x
    #define _PHA                0x48
    #define _PLB                0xAB
    
    /// move this to an overworld related header
    // stock object (overworld) types, listed out as map16 values
    // These are map16 values that have special significance to the game engine
    // They are treated as protected values that cannot be deallocated during Map16 editing as well.
    typedef enum
    {
        stock_empty     = 0x0034,
        stock_bush      = 0x0036,
        stock_bush2     = 0x072A,
        
        stock_rock_sl   = 0x020F,
        stock_rock_sh   = 0x0239,
        
        stock_peg       = 0x021B,
        
        stock_rock_bl0  = 0x036D,
        stock_rock_bl1  = 0x036E,
        stock_rock_bl2  = 0x0374,
        stock_rock_bl3  = 0x0375,
        
        stock_rock_bh0  = 0x023B,
        stock_rock_bh1  = 0x023C,
        stock_rock_bh2  = 0x023D,
        stock_rock_bh3  = 0x023E,
        
        stock_warp      = 0x0212,
        
        stock_rockPile0 = 0x0226,
        stock_rockPile1 = 0x0227,
        stock_rockPile2 = 0x0228,
        stock_rockPile3 = 0x0229,
        
        stock_sign      = 0x0101,
        
        stock_grass     = 0x037E,
        
        stock_jumpPoint = 0x066B,
        
        stock_stump0    = 0x008E,
        stock_stump1    = 0x008F,
        stock_stump2    = 0x0098,
        stock_stump3    = 0x0099,
        
        stock_burrow0   = 0x0918,
        stock_burrow1   = 0x0919,
        stock_burrow2   = 0x091A,
        stock_burrow3   = 0x091B
        
    } stockMap16;

#endif
