    
#ifndef CONTAINERS_HEADER_GUARD

    #include "containers.h"
    
#endif

// If this .cpp file is being included in a header
#ifndef CONTAINERS_SOURCE_GUARD
    
    #define CONTAINERS_SOURCE_GUARD
    
namespace Containers
{
    
// ===============================================================
    
    // _______________________________________________________
    template<typename T>
    Slot<T>::Slot() : m_contents(NULL)
    {
        
    }
    
    // _______________________________________________________
    template<typename T>
    Slot<T>::Slot(T arg) : m_contents(NULL)
    {
        Set(arg);
    }
    
    // _______________________________________________________
    template<typename T>
    Slot<T>::Slot(Slot<T> &rhs) : m_contents(NULL)
    {
        if(this != &rhs)
        {
            if(rhs.IsEmpty())
            {
                Clear();
            }
            else
            {
                Set(rhs.Get());
            }
        }
    }
    
    // _______________________________________________________
    // Copy constructor that takes a constant reference.
    template<typename T>
    Slot<T>::Slot(Slot<T> const &rhs) : m_contents(NULL)
    {
        if(this != &rhs)
        {
            if(rhs.IsEmpty())
            {
                Clear();
            }
            else
            {
                Set(rhs.Get());
            }
        }
    }
    
    // _______________________________________________________
    template<typename T>
    Slot<T>& Slot<T>::operator = (Slot<T> &rhs)
    {
        if(this != &rhs)
        {
            if( rhs.IsEmpty() )
            {
                Clear();
            }
            else
            {
                Set(rhs.Get());
            }
        }
        
        return (*this);
    }
    
    // _______________________________________________________
    // Assignment operator that accepts a constant argument
    template<typename T>
    Slot<T>& Slot<T>::operator = (Slot<T> const &rhs)
    {
        if(this != &rhs)
        {
            if( rhs.IsEmpty() )
            {
                Clear();
            }
            else
            {
                Set(rhs.Get());
            }
        }
        
        return (*this);
    }
    
    // _______________________________________________________
    // Is this necessary or automatically done already?
    template<typename T>
    Slot<T>& Slot<T>::operator = (T arg)
    {
        Set(arg);
        
        return (*this);
    }
    
    // _______________________________________________________
    template<typename T>
    Slot<T>::~Slot()
    {
        Clear();
    }
    
    // _______________________________________________________
#if 1
    template<typename T>
    void Slot<T>::Set(T arg)
    {
        if( IsEmpty() )
        {
            // Need to generate a pointer to reference the new value.
            m_contents = new T;
        }
        
        (*m_contents) = arg;
    }
#endif
    
#if 0
    // _______________________________________________________
    template<typename T>
    void Slot<T>::Set(T const &arg)
    {
        if( IsEmpty() )
        {
            // Need to generate a pointer to reference the new value.
            contents = new T;
        }
        
        (*contents) = arg;
    }
#endif
    
    // _______________________________________________________
    template<typename T>
    void Slot<T>::Clear()
    {
        if( !IsEmpty() )
        {
            delete m_contents;
            
            m_contents = NULL;
        }
    }
    
    // _______________________________________________________
    template<typename T>
    bool Slot<T>::IsEmpty()
    {
        return (m_contents == NULL);
    }
    
    // _______________________________________________________
    template<typename T>
    bool Slot<T>::IsEmpty() const
    {
        return (m_contents == NULL);
    }
    
    // _______________________________________________________
    template<typename T>
    T& Slot<T>::operator * ()
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
    // _______________________________________________________
    template<typename T>
    T Slot<T>::operator * () const
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
    // _______________________________________________________
    // These sure are some odd semantics... wonder why I'd use ->
    // to get a pointer?
    template<typename T>
    T* Slot<T>::operator -> () const
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return m_contents;
    }
    
    // _______________________________________________________
    // Can implicitly convert back to the contained type.
    template<typename T>
    Slot<T>::operator T() const
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
    // _______________________________________________________
    // Implicitly convert to a boolean value of true when the
    // slot is nonempty, and false when it's empty.
    template<typename T>
    Slot<T>::operator bool() const
    {
        return (!IsEmpty());
    }
    
    // _________________________________________________________________________
    template<typename T>
    T Slot<T>::Get()
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
    // _________________________________________________________________________
    template<typename T>
    T Slot<T>::Get() const
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
    // _________________________________________________________________________
    template<typename T>
    T& Slot<T>::GetRef()
    {
        if( IsEmpty() )
        {
            throw exception();
        }
        
        return (*m_contents);
    }
    
// =============================================================================
    
}

#endif
