    
    #include <stdio.h>
    
    #include <exception>
    #include <list>
    #include <string>
    #include <memory>
    
    // Temporary include. The ToString() functionality of Matrix currently
    // demands it..........
    #include <typeinfo>
    
    #include "inttypes.h"

#ifndef CONTAINERS_HEADER_GUARD

    #define CONTAINERS_HEADER_GUARD

namespace Containers
{
    using std::exception;
    using std::list;
    using std::pair;
    using std::string;
    
    using std::shared_ptr;
    
// =============================================================================
    
    template<typename T>
    class Slot
    {
        
    private:
        
        T *m_contents;
        
    public:
        
        Slot();
        
        Slot(T arg);
        
        Slot(Slot<T> &rhs);
        Slot(Slot<T> const &rhs);
        
        Slot<T>& operator = (Slot<T> &rhs);
        Slot<T>& operator = (Slot<T> const &rhs);
        
        Slot<T>& operator = (T arg);
        
        ~Slot();
        
    #if 1
        void Set(T arg);
    #endif
        
    #if 0
        void Set(T const &arg);
    #endif
        
        void Clear();
        
        bool IsEmpty();
        bool IsEmpty() const;
        
        T& operator * ();
        T  operator * () const;
        
        T* operator -> () const;
        
        operator T() const;
        
        operator bool() const;
        
        T Get();
        T Get() const;
        
        T& GetRef();
    };
    
// =============================================================================
    
    template<class T>
    class Ticket
    {
    protected:
        
        typedef shared_ptr< Slot<T> > shared_slot;
        
        shared_slot m_act;
        
    public:
        
        Ticket()
        
        : m_act()
        {
        }
        
        Ticket(shared_slot p_act)
        
        : m_act(p_act)
        {
            
        }
        
        // _____________________________________________________________________
        virtual ~Ticket()
        {
            
        }
        
        // _____________________________________________________________________
        Ticket(Ticket const &t)
        
        : m_act(t.m_act)
        {
        }
        
        // _____________________________________________________________________
        Ticket<T>& operator = (Ticket<T> const &t)
        {
            if( &t != this )
            {
                m_act = t.m_act;
            }
            
            return (*this);
        }
        
        // _____________________________________________________________________
        bool Valid() const
        {
            return (m_act->IsEmpty() == false);
        }
        
        // _____________________________________________________________________
        T Redeem() const
        {
            return (*m_act);
        }
    };

// =============================================================================

    template<class T>
    class Show
    {
        
    protected:
        
        typedef shared_ptr< Slot<T> > shared_slot;
        
        shared_slot m_act;
        
    public:
        
        // _____________________________________________________________________
        //! \brief      Provides a Show that is a handle to a default
        //!             constructed object of type T.
        Show()
        
        : m_act( shared_slot( new Slot<T>( T() ) ) )
        {
            
        }
        
        // _____________________________________________________________________
        Show(T p_act)
        
        : m_act( shared_slot( new Slot<T>(p_act) ) )
        {
        }
        
        // _____________________________________________________________________
        Show(Show const &p_s)
        
        : m_act(p_s.m_act)
        {
            
        }
        
        // _____________________________________________________________________
        virtual ~Show()
        {
            
        }
        
        // _____________________________________________________________________
        // Used for 
        Show<T>& operator = (Show const &p_s)
        {
            if(this != &p_s)
            {
                m_act = p_s.m_act;
            }
            
            return (*this);
        }
        
        // _____________________________________________________________________
        Show<T>& operator = (T &t)
        {
            m_act->Set(t);
            
            return (*this);
        }
        
        // _____________________________________________________________________
        // Derive a ticket from this object.
        Ticket<T> Issue()
        {
            shared_slot s = m_act;
            
            return Ticket<T>(s);
        }
        
        // _____________________________________________________________________
        void Nullify()
        {
            m_act->Clear();
        }
        
        // _____________________________________________________________________
        T Get()
        {
            return (*m_act);
        }
        
        // _____________________________________________________________________
        T& GetRef()
        {
            Slot<T> *p = m_act.get();
            
            // -----------------------------
            
            return ( p->GetRef() );
        }
    };
    
// =============================================================================

    template <class T>
    class Matrix
    {
    public:
        
        int m_rows;
        int m_columns;
        
        T **m_contents;
        
        // _____________________________________________________________________
        typedef list< Matrix<T> > MatrixList;
        
        typedef pair< Matrix<T>, Matrix<T> > MatrixPair;
        
        typedef list< MatrixPair > MatrixPairList;
        
        // _____________________________________________________________________
        Matrix(int rows                = 0,
               int columns             = 0,
               T const * const array1D = NULL)
        
        : m_rows(0),
          m_columns(0),
          m_contents(nullptr)
        {
            int i = 0, j = 0;
            int k = 0;
            
            // ------------------------------
            
            Init(rows,
                 columns);
            
            if(array1D != NULL)
            {
                for(i = 0; i < m_rows; ++i)
                {
                    for(j = 0; j < m_columns; ++j)
                    {
                        m_contents[i][j] = array1D[k];
                        
                        k += 1;
                    }
                }
            }
        }
        
        // _____________________________________________________________________
        Matrix(int     p_rows,
               int     p_columns,
               list<T> p_values)
        
        : m_rows(0),
          m_columns(0),
          m_contents(nullptr)
        {
            Init(p_rows,
                 p_columns);
            
            if(1)
            {
                typename list<T>::iterator it     = p_values.begin();
                typename list<T>::iterator end_it = p_values.end();
                
                // ------------------------------
                
                for(int i = 0; i < p_rows; ++i)
                {
                    if(it == end_it)
                    {
                        break;
                    }
                    
                    for(int j = 0; j < p_columns; ++j)
                    {
                        (*this)(i, j) = (*it);
                        
                        ++it;
                    }
                }
            }
        }

        // _____________________________________________________________________
        Matrix(const Matrix<T> &m)
        
        : m_rows(0),
          m_columns(0),
          m_contents(nullptr)
        {
            Init(m.m_rows,
                 m.m_columns);
            
            if( NonEmpty() )
            {
                for(int i = 0; i < m.m_rows; ++i)
                {
                    for(int j = 0; j < m.m_columns; ++j)
                    {
                        (*this)(i, j) = m(i, j);
                    }
                }
            }
        }
        
    // =========================================================
        
        Matrix<T>& operator = (Matrix<T> const &rhs)
        {
            if(&rhs != this)
            {
                Init(rhs.m_rows,
                     rhs.m_columns);
                
                if( NonEmpty() )
                {
                    for(int i = 0; i < rhs.m_rows; ++i)
                    {
                        for(int j = 0; j < rhs.m_columns; ++j)
                        {
                            (*this)(i,j) = rhs(i,j);
                        }
                    }
                }
            }
            
            return (*this);
        }
        
        // _____________________________________________________________________
        int Size()
        {
            return (m_rows * m_columns);
        }
        
        // _____________________________________________________________________
        virtual ~Matrix()
        {
            Nullify();
        }
        
        // _____________________________________________________________________
        void Init(int p_rows,
                  int p_columns)
        {
            // Delete the existing allocation of rows and columns, if we have
            // any.
            if(Size() > 0)
            {
                Nullify();
            }
            
            // Set the new matrix dimensions now and allocate new memory to
            // accomodate it.
            m_rows    = (p_rows    < 0) ? 0 : p_rows;
            m_columns = (p_columns < 0) ? 0 : p_columns;
            
            if( NonEmpty() )
            {
                m_contents = new T*[m_rows];
                
                for(int i = 0; i < m_rows; ++i)
                {
                    m_contents[i] = new T[m_columns];
                }
            }
        }
        
        // _____________________________________________________________________
        void Nullify()
        {
            if(NonEmpty())
            {
                int i = 0;
                
                // -----------------------------
                
                for(i = 0; i < m_rows; ++i)
                {
                    delete[] m_contents[i];
                    
                    m_contents[i] = NULL;
                }
                
                delete[] m_contents;
            }
        }
        
        // _____________________________________________________________________
        bool NonEmpty() const
        {
            if(m_rows < 0)
            {
                return false;
            }
            
            if(m_columns < 0)
            {
                return false;
            }
            
            return true;
        }
        
        // _____________________________________________________________________
        T & operator() (int row, int column)
        {
            if( (row >= m_rows) || (row < 0) )
            {
                throw exception();
            }
            else if( (column >= m_columns) || (column < 0) )
            {
                throw exception();
            }
            
            return m_contents[row][column];
        }
        
        // _____________________________________________________________________
        // For accessing elements of a constant matrix.
        T operator() (int row, int column) const 
        {
            if( (row >= m_rows) || (row < 0) )
            {
                throw exception();
            }
            else if( (column >= m_columns) || (column < 0) )
            {
                throw exception();
            }
            
            return m_contents[row][column];
        }
        
        // _____________________________________________________________________
        Matrix<T> operator() (int row,
                              int column,
                              int rows,
                              int columns) const
        {
            // grabs a sub matrix
            
            // ----------------------------------------
            
            if( (row >= m_rows) || (row < 0) )
            {
            }
            else if( (column >= m_columns) || (column < 0) )
            {
            }
            else if( ( (row + rows) > m_rows ) || (rows < 0) )
            {
            }
            else if( ( (column + columns) > m_columns) || (columns < 0) )
            {
            }
            else
            {
                Matrix<T> m(rows, columns);
                
                // ------------------------------
                
                for(int i = 0; i < rows; ++i)
                {
                    for(int j = 0; j < columns; ++j)
                    {
                        m(i, j) = (*this)(i + row, j + column);
                    }
                }
                
                return m;
            }
            
            // return an empty matrix if the event of an exception
            return Matrix<T>();
        }
        
        // _____________________________________________________________________
        bool operator == (Matrix<T> const &m) const
        {
            int i = 0, j = 0;
            
            // -------------------------------
            
            if(m.m_rows != m_rows)
            {
                return false;
            }
            
            if(m.m_columns != m_columns)
            {
                return false;
            }
            
            for(i = 0; i < m.m_rows; ++i)
            {
                for(j = 0; j < m.m_columns; ++j)
                {
                    if( (*this)(i, j) != m(i, j) )
                    {
                        return false;
                    }
                }            
            }
            
            return true;
        }
        
        // _____________________________________________________________________
        bool operator != (Matrix<T> const &m)
        {
            return !( (*this) == m );
        }        
        
        // _____________________________________________________________________
        bool operator < (Matrix<T> const &n) const
        {
            int i = 0, j = 0;
            
            // ---------------------------
            
            if( (m_rows * m_columns) < (n.p_rows * n.p_columns) )
            {
                return true;
            }
            
            /// need a weak criterion!!!!
            
            return false;
        }
    
        // _____________________________________________________________________
        static Matrix<T> ConcatRight(Matrix<T> const &m1,
                                       Matrix<T> const &m2)
        {
            if(m1.m_rows != m2.m_rows)
            {
                throw exception();
                
                return Matrix<T>(0, 0);
            }
            else
            {
                int i = 0, j = 0;
                
                Matrix<T> m3 = Matrix<T>(m1.m_rows,
                                         m1.m_columns + m2.m_columns);
                
                // --------------------------------------------
                
                // copy the first matrix to the far left of the new matrix
                for(i = 0; i < m3.m_rows; ++i)
                {
                    for(j = 0; j < m1.m_columns; ++j)
                    {
                        m3(i, j) = m1(i, j);
                    }
                }
                
                // and copy the second directly to the right of the first's position 
                for(i = 0; i < m3.m_rows; ++i)
                {
                    for(j = 0; j < m2.m_columns; ++j)
                    {
                        m3(i, j + m1.m_columns) = m2(i, j);
                    }
                }
                
                return m3;
            }
        }
        
        // _____________________________________________________________________
        static Matrix<T> ConcatDown(Matrix<T> const &m1,
                                    Matrix<T> const &m2)
        {
            if(m1.m_columns != m2.m_columns)
            {
                throw exception();
                
                return Matrix<T>(0, 0);
            }
            else
            {
                int i = 0, j = 0;
                
                Matrix<T> m3 = Matrix<T>(m1.m_rows + m2.m_rows,
                                         m1.m_columns);
                
                // --------------------------------------------
                
                // copy the top matrix to the upper portion of the new matrix
                for(i = 0; i < m1.m_rows; ++i)
                {
                    for(j = 0; j < m1.m_columns; ++j)
                    {
                        m3(i, j) = m1(i, j);
                    }
                }
                
                // copy in the second matrix directly below the first's position.
                for(i = 0; i < m2.m_rows; ++i)
                {
                    for(j = 0; j < m2.m_columns; ++j)
                    {
                        m3(i + m1.m_rows, j) = m2(i, j);
                    }
                }
                
                return m3;
            }
        }
        
        // _____________________________________________________________________
        string ToString()
        {
            string s;
            
            // -----------------------
            
            for(int i = 0; i < m_rows; ++i)
            {
                for(int j = 0; j < m_columns; ++j)
                {
                    char rep[100];
                    
                    if( typeid(T) == typeid(uint8_t) )
                    {
                        sprintf(rep, "%d", m_contents[i][j]);
                        s += rep;
                    }
                    else if( typeid(T) == typeid(uint16_t) )
                    {
                        sprintf(rep, "%d", m_contents[i][j]);
                        s += rep;
                    }
                    else if( typeid(T) == typeid(uint32_t) )
                    {
                        sprintf(rep, "%d", m_contents[i][j]);
                        s += rep;
                    }
                    else
                    {
                        // assume there's a type conversion to string
                        // for other classes
                        s += m_contents[i][j];
                    }
                    
                    // conditions for putting a comma and space after the text
                    if( (j < (m_columns - 1) || (i < (m_rows - 1) ) ) )
                    {
                        s += ", ";
                    }
                }
                
                if(i < (m_rows - 1) )
                {
                    s += "\n";
                }
            }
            
            return s;
        }

        // _____________________________________________________________________
        int Count(Matrix<T> const &m)
        {
            int count = 0;
            
            // -----------------------------
            
            if( m.m_rows > m_rows )
            {
                throw exception();
            }
            else if( m.m_columns > m_columns )
            {
                throw exception();
            }
            
            for(int i = 0; i < (m_rows - m.m_rows) + 1; ++i)
            {
                for(int j = 0; j < (m_columns - m.m_columns) + 1; ++j)
                {
                    bool allMatch = true;
                    
                    // ------------------------------
                    
                    for(int i2 = 0; (i2 < m.m_rows) && (allMatch == true); ++i2)
                    {
                        for(int j2 = 0; j2 < m.m_columns; ++j2)
                        {
                            if( m_contents[i + i2][j + j2] != m(i2, j2) )
                            {
                                allMatch = false;
                                
                                break;                                
                            }                            
                        }                        
                    }
                    
                    if(allMatch)
                    {
                        count += 1;
                    }                      
                }
            }
            
            return count;
        }
        
        // _____________________________________________________________________
        int Coverage(Matrix<T> const &m)
        {
            int count = 0;
            
            Matrix<bool> matches(m_rows, m_columns);
            
            // -----------------------------
            
            if( m.m_rows > m_rows )
            {
                throw exception();
            }
            else if( m.m_columns > m_columns )
            {
                throw exception();
            }
            
            for(int i = 0; i < m_rows; ++i)
            {
                for(int j = 0; j < m_rows; ++j)
                {
                    matches(i, j) = false;
                }
            }
            
            for(int i = 0; i < (m_rows - m.m_rows) + 1; ++i)
            {
                for(int j = 0; j < (m_columns - m.m_columns) + 1; ++j)
                {
                    bool allMatch = true;
                    
                    // ---------------------------
                    
                    for(int i2 = 0; (i2 < m.m_rows) && (allMatch == true); ++i2)
                    {
                        for(int j2 = 0; j2 < m.m_columns; ++j2)
                        {
                            if( (*this)(i + i2, j + j2) != m(i2, j2) )
                            {
                                allMatch = false;
                                
                                break;
                            }
                        }
                    }
                    
                    if(allMatch)
                    {
                        for(int i2 = 0; i2 < m.m_rows; ++i2)
                        {
                            for(int j2 = 0; j2 < m.m_columns; ++j2)
                            {
                                matches(i + i2, j + j2) = true;
                            }                        
                        }
                    }                    
                }
            }
            
            // finally, count all the entries that matched
            for(int i = 0; i < m_rows; ++i)
            {
                for(int j = 0; j < m_rows; ++j)
                {
                    if(matches(i, j))
                    {
                        count += 1;
                    }
                }
            }
            
            return count;
        }
        
        // _____________________________________________________________________
        Matrix<T> GetColumn(int c)
        {
            return (*this)(0, c, m_rows, 1);
        }        
        
        // _____________________________________________________________________
        Matrix<T> GetRow(int r)
        {
            return (*this)(r, 0, 1, m_columns);
        }        
        
        // _____________________________________________________________________
        bool IsCyclic()
        {
            bool h_cyclic = false;
            bool v_cyclic = false;
            
            string s = ToString();
            
            // -------------------------
            
            for(int r = 1; r < m_rows; ++r)
            {
                if( (*this).GetRow(0) == (*this).GetRow(r) )
                {
                    h_cyclic = true;
                    
                    for(int r2 = 1; (r + r2) < m_rows; ++r2)
                    {
                        if( (*this).GetRow(r2) != (*this).GetRow(r + r2) )
                        {
                            h_cyclic = false;
                            break;   
                        }           
                    }
                    
                    if(h_cyclic)
                    {
                        break;
                    }
                }
            }
            
            for(int c = 1; c < m_columns; ++c)
            {
                if( (*this).GetColumn(0) == (*this).GetColumn(c) )
                {
                    v_cyclic = true;
                    
                    for(int c2 = 1; (c + c2) < m_columns; ++c2)
                    {
                        if( (*this).GetColumn(c2) != (*this).GetColumn(c + c2) )
                        {
                            v_cyclic = false;
                            break;   
                        }           
                    }
                    
                    if(v_cyclic)
                    {
                        break;
                    }
                }
            }
            
            return (h_cyclic | v_cyclic);        
        }
        
    };
    
// =============================================================================
    
}

#endif

#ifndef CONTAINERS_SOURCE_GUARD
    
    #include "containers.cpp"
    
#endif


