    
    #include "dungtypes.h"
    
    #include <memory>

namespace Dungeon
{

// =============================================================================
    
    Room::Room()
    {   
        
    }

// =============================================================================

    Header::Header()
    
    : m_bg0_props(0),
      m_collision(0),
      m_lightsOut(0),
      m_paletteIndex(0),
      m_bgGfx(0),
      m_spriteGfx(0x40),
      m_effect(0),
      m_tag_1(0),
      m_tag_2(0),
      m_warpPlane(0),
      m_staircasePlane1(0),
      m_staircasePlane2(0),
      m_staircasePlane3(0),
      m_staircasePlane4(0),
      m_warpRoom(0)
    {
        for(int i = 0; i < 4; ++i)
        {
            m_staircaseRooms[i] = 0;
        }
        
        return;
    }

// =============================================================================

    Header::Header(Buffer *data)
    {
        Initialize(data);
        
        return;
    }
    
// =============================================================================

    void Header::Initialize(Buffer *data)
    {
        uint8_t currentByte;
        
        // -----------------------------------
        
        currentByte = data->GetByte(0);
        
        m_bg0_props       = (currentByte >> 5) & 0x07;
        m_collision       = (currentByte >> 2) & 0x07;
        m_lightsOut       = (currentByte)      & 0x01;
        
        m_paletteIndex    = data->GetByte(1);
        m_bgGfx           = data->GetByte(2);
        m_spriteGfx       = data->GetByte(3) + 0x40;
        m_effect          = data->GetByte(4);
        m_tag_1           = data->GetByte(5);
        m_tag_2           = data->GetByte(6);
        
        currentByte           = data->GetByte(7);
        
        // The modulo 3 (% 3) is because 3 is not actually a valid value.
        m_warpPlane       = ((currentByte >> 0) & 0x03) % 3;
        m_staircasePlane1 = ((currentByte >> 2) & 0x03) % 3;
        m_staircasePlane2 = ((currentByte >> 4) & 0x03) % 3;
        m_staircasePlane3 = ((currentByte >> 6) & 0x03) % 3;
        
        m_staircasePlane4 = (data->GetByte(8) & 0x03 ) % 3;
        
        m_warpRoom        = data->GetByte(9);
        
        for(int i = 0; i < 4; ++i)
        {
            m_staircaseRooms[i] = data->GetByte(10 + i);
        }
        
        return;
    }

// =============================================================================

    uint8_t Header::GetStaircaseRoom(int i) const
    {
        if( (0 <= i) && (i <= 4) )
        {
            return m_staircaseRooms[i];
        }
        
        return 0;
    }

// =============================================================================

    Object::Object(objType p_type, uint32_t p_data)
    
    : m_disabled(false)
    {
        m_type = p_type;
        
        switch(p_type)
        {
            default:
            case Object::type1_1:
                
                m_addr     = ( (p_data & 0x00FC) >> 1) | ( (p_data & 0xFC00) >> 3);
                                    
                m_arg1     = (p_data >> 0)  & 0x03;
                m_arg2     = (p_data >> 8)  & 0x03;
                
                m_routine  = (p_data >> 16) & 0xFF;
                
                break;
                
            case Object::type1_2:
                
                // I really hate subtype 2's data format
                m_addr     = (p_data & 0x00F000) >> 11; // (e bits)
                m_addr    |= (p_data & 0x000003) << 5;  // (b bits)
                m_addr    |= (p_data & 0xC00000) >> 15; // (f bits)
                m_addr    |= (p_data & 0x000F00) << 1;  // (c bits) (referencing Zelda_ROM.log)
                
                m_arg1     = 0;
                m_arg2     = 0;
                
                m_routine  = (p_data & 0x3F0000) >> 16;
                
                switch(m_routine)
                {
                    default:
                        
                        m_routine = m_routine;
                        
                        break;
                        
                    case 0x3A:
                        
                        m_routine = m_routine;
                        
                        break;
                }
                
                break;
                
            case Object::type1_3:
            
                m_addr     =   ( (p_data & 0x00FC) >> 1 )
                             | ( (p_data & 0xFC00) >> 3 );
                
                m_arg1     = 0;
                m_arg2     = 0;
                
                m_routine  = (p_data & 0x070000) >> 12; // (e bits)
                m_routine |= (p_data & 0x000300) >> 6;  // (c bits)
                m_routine |= (p_data & 0x000003) >> 0;  // (a bits)
                
                switch(m_routine)
                {
                    default:
                    
                        m_routine = m_routine;
                        
                        break;
                        
                    case 0x1E:
                    case 0x1F:
                    case 0x20:
                    case 0x21:

                        m_routine = m_routine;
                        
                        break;
                }
                
                break;
        }
        
        
        m_x = Tilemap::ToX(m_addr);
        m_y = Tilemap::ToY(m_addr);
    }
    
    Object::Object(Object const &p_o)
    
    : m_disabled(p_o.m_disabled),
      m_routine(p_o.m_routine),
      m_addr(p_o.m_addr),
      m_arg1(p_o.m_arg1),
      m_arg2(p_o.m_arg2),
      m_width(p_o.m_width),
      m_height(p_o.m_height),
      /// m_layer(o.m_layer), /// not sure whether to enable this
      m_x(p_o.m_x),
      m_y(p_o.m_y),
      m_type(p_o.m_type)
    {
    
    
    }

// =============================================================================

    State::State(Entrance const &entr,
                 Room const     &room,
                 ObjectList     &layout)
    {
        int layer = 0;
        
        Object obj;
        ObjectList list;
        ObjectList::iterator it;
        
        // ------------------------------------
        
        for(it = layout.begin(); it != layout.end(); it++)
        {
            obj = (*it);
            
            ParseObject(obj);
        }
        
        for(layer = 0; layer < 3; layer += 1)
        {
            list = room.objects[0].type1;
            
            for( it = list.begin(); it != list.end(); it++)
            {
                obj = (*it);
                
                ParseObject(obj);
            }
        }
    }
    
// =============================================================================

    void State::ParseObject(Object &p_obj)
    {
        using std::exception;
        
        switch(p_obj.m_type)
        {
            case Object::type1_1:
            
                /// problem: we don't have a rom object in this class...
                /// or namespace even...
                /// Simulate(obj->routine);
            
                break;
            
            case Object::type1_2:
                
                
                
                break;
            
            case Object::type1_3:
                
                
                
                break;
            
            default:
            
                throw exception( (const exception&) "ParseObjectException");
                break;
        }
    }

// =============================================================================

}

namespace Thorn
{
    
// =============================================================================

    Obj_Def test_obj_def
    {
        {
            { string("b0"), block_ref }
        },
        {
            
        }
    };
    
    Obj_Param test =
    {
        "b0",
        block_ref
    };
    
    Block obj_1_00
    {
        2, 2,
        {
            0x0000,
            0x0001,
            0x0002,
            0x0003
        }
    };
    
    void TestSharedPtr()
    {
        using std::shared_ptr;
        
        shared_ptr<Block> sharedBlock;
        
        {
            Block *b = new Block(2, 2, { 0, 1, 2, 3} );
            
            delete b;
            
            shared_ptr<Block> sharedBlock2(new Block(2, 2, { 0, 1, 2, 3} ) );
        }
        
        sharedBlock.reset( new Block(2,
                                     2,
                                     {
                                         0x0000,
                                         0x0001,
                                         0x0002,
                                         0x0003
                                     } ) );
        
    }
    
    std::set<Block> a;
    
    void Durp()
    {
        std::set<Block> b;
        
        IndirectBlock c
        {
            0x1000,
            2,
            2,
            {
                0x0000,
                0x0001,
                0x0002,
                0x0003
            }
        };
        
        a.swap(b);
    }
}

// =============================================================================

