﻿
    #include "inttypes.h"
    
    #include "constants.h"
    #include "buffer.h"

#ifndef EXPANDINFO_H

    #define EXPANDINFO_H

    class ExpandInfo
    {
        
    public:
        
        static const uint32_t expand_version;
        
        // Flag indicating whether the file has an extra 0x200 bytes
        bool m_hasHeader;
        
        // pointer for the header's data so we can logically separate rom and header
        Buffer *m_headerBuf;
        
        // copy of the ROM provided from ZeldaGame or otherwise
        Buffer *m_rom;
        
        // buffer used for serialization of expansion info
        Buffer *m_data;
        
        // defaults to "EXPAND_INFO"
        char m_designation[12];
        
        struct
        {
            uint32_t version;
            uint32_t size;
            
        } m_header;
        
        struct
        {
            uint32_t headerOffset;
            uint32_t headerBank;
            uint32_t objOffset;
            uint32_t entranceOffset;
            uint32_t chestOffset;
            uint32_t numChests;
            uint32_t spriteOffset;
            uint32_t spriteBanks;
            uint32_t layoutOffset; // Offset in the rom where the dungeon layout (template) object data is at.
        
        } m_dungeon;
        
        struct
        {
            uint32_t offset;  // The offset where the pointer table for all the dialogue is at.
            uint32_t numMsgs; // Number of dialogue messages being stored in the rom.
        
        } m_text;
        
        struct
        {
            uint32_t numParts;   // Number of parts used in the game (not implemented yet)
            
            uint32_t holeOffset; // CPU pointer to the beginning of the overworld hole data.
            uint32_t numHoles;   // The number of hole markers that are on the overworld.
            
            uint32_t entrOffset; // CPU pointer to the beginning of the overworld entrance data.
            uint32_t numEntr;    // The number of entrance markers (not the same as the number of entrances to the dungeons) on the overworld.
            
            uint32_t exitOffset; // location where the base exit data can be found.
            uint32_t numExits;   // Number of exits in the game.
            
            uint32_t itemOffset; // CPU pointer to the table that has all the overworld item data.
            
            uint32_t sprOffset0; // CPU pointer to the sprite data table for parts 0 and 1
            uint32_t sprOffset2; // CPU pointer to the sprite data table for part 2
            uint32_t sprOffset3; // CPU pointer to the sprite data table for part 3
            
            uint32_t map32To16UL;    // CPU pointer to the map32 to map16 conversion array (upper left)
            uint32_t map32To16UR;    // CPU pointer to the map32 to map16 conversion array (upper right)
            uint32_t map32To16LL;    // CPU pointer to the map32 to map16 conversion array (lower left)
            uint32_t map32To16LR;    // CPU pointer to the map32 to map16 conversion array (lower right)
            
        } m_overworld;
        
        struct
        {
            uint32_t sprOffset;       // CPU pointer to the sprite graphics pointer array
            uint32_t bgOffset;        // CPU pointer to the background graphics pointer array
            
        } m_graphics;
        
    public:
    
        ExpandInfo(std::ifstream &zeldaFile);
        ~ExpandInfo();
        
        void Initialize();
        void Reinitialize();
        void Serialize();
        
        bool CheckForExpansion();
        bool Deserialize(uint32_t offset);
        
    };

#endif