
    // #include "direct.h"

    /// #include "Black Magic.h"
    #include "globals.h"
    #include "strings.h"
    #include "core.h"

    #include "compresch_lttp.h"

// ===============================================================

    typedef enum 
    {
        mode_no_compress,
        mode_byte_repeat,
        mode_word_repeat,
        mode_increment,
        mode_copy

    } comp_mode;

// ===============================================================

    // CompressStandard is an adaptation of the compression algorithm found in Hyrule Magic.
    // For my part, this is the only piece of code that was taken from HM, and I don't even use it currently.
    // Compresch, Disch's compression library has pretty much tied up all my loose ends on the compression front.

    bufPtr CompressStandard(bufPtr srcBuf, int flag)
    {
        if(!srcBuf)
            return NULL;

        // allocate a 2^12 sized buffer
        bufPtr destBuf = CreateBuffer(0x1000); 

        if(!destBuf)
            return NULL;

        destBuf->wOffset = 0;

        int i = 0, j = 0, k = 0;
    
        int byte      = 0;
        int highByte  = 0;
        int remainder = 0;
        int word      = 0;
        int identical = 0;
        int oldSize = srcBuf->length;
    
        comp_mode p = mode_no_compress;
    
        int n = 0;
        int x = 0, q = 0, r;
    
        // flag is an endianness flag as follows
        // flag ?  bigEndian() : littleEndian();
        // hence default behavior is littleEndian

        // int *size; this is the return size of the compressed data
        // after the routine has finished
    
	    for(i = 0; i < oldSize; ) 
	    {
		    byte = GetByte(srcBuf, i); // grab a char from the buffer.
		    
		    k = 0;
		    
            // r = the same logical value (0 or 1) as q, but not the same value necesarily.
    	    r = !!q; 
        
            // looking at the bytes from the start of the array up until the byte at i for some reason...
		    for(j = 0; j < i - 1; j++) 
            {
                // 
                if( GetByte(srcBuf, j) == byte) 
			    {
				    for(n = j; n < oldSize; n++) 
				    {	
					    if( GetByte(srcBuf, n) != GetByte(srcBuf, n + i) )
						    break;
                    }
			    
                    if(n > k) 
                        k = n, word = j;
                }
            }
 
            identical = 0;

            // get the number of bytes starting at i that are identical to
            // the byte at i
            for( j = (i + 1); j < oldSize; j++ ) 
            {
                if( GetByte(srcBuf, j) != byte ) 
                    break;
                
                identical++;
            }
    
            if( identical > (1 + r) ) 
                p = mode_byte_repeat; 
            else 
            {
                // if looking for repeating bytes failed, we're going to try
                // to find a sequence of repeating words (16-bit) instead

                word = Get2Bytes(srcBuf, i);
                identical = 0;

                for(j = i + 2; j < oldSize; j++) 
                {            
                    // while words are matched, the total number of actual bytes
                    // matched is recorded, not the number of words
                    if(Get2Bytes(srcBuf, j) != word)
                    {
                        if(GetByte(srcBuf, j) == byte)
                            identical++;
                    
                        break;
                    }
                    else
                        identical += 2;
                }
        
                if( identical > (2 + r) ) 
                    p = mode_word_repeat; 
                else 
                {
                    // well that didn't work so we're going to try to
                    // find an incrementing byte series

                    // MON is iffy on how well this will actually work...
                    // due to how the actual algorithm decompresses stuff...

                    identical = 0;

                    for(j = (i + 1); j < oldSize; j++) 
                    {
                        if(GetByte(srcBuf, j) != (byte + j) ) 
                            break;

                        identical++;
                    }
            
                    if( identical > (1 + r) ) 
                        p = mode_increment; 
                    else 
                        p = mode_no_compress;
                }
            }
		    
            // what the fuck is this
            if( (k > r + 3) && (k > n + (p % 2)) ) 
            {
                p = mode_copy;
                n = k;
            }
    
            if(!p) 
                q++, i++; 
            else 
            {
                if(q) 
                {
                    q--; 

                    if(q > 0x1F) 
                        PutByte(destBuf, 0xE0 + (q >> 8) );
            
                    PutByte(destBuf, q);
                    q++;

                    CopyBuffer(destBuf, srcBuf, 
                               destBuf->wOffset,      i - q, 
                               q);

                    destBuf->wOffset += q;
                    q = 0;
                }

                i += identical;
                identical--;
                        
                if(identical > 0x1F)
                {
                    PutByte(destBuf, 0xE0 + (identical >> 8) + (p << 2));
                    PutByte(destBuf, identical);
                }
                else 
                    PutByte(destBuf, ((p << 5)+ identical));

                switch(p) 
                {
                case mode_byte_repeat: 
                case mode_increment: 

                    PutByte(destBuf, byte); 
                    break;
        
                case mode_word_repeat: 

                    Put2Bytes(destBuf, word);
                    break;

                case mode_copy: 
                    if(flag) 
                    { 
                        PutByte(destBuf, word >> 8);
                        PutByte(destBuf, word);
                    }
                    else 
                    { 
                        Put2Bytes(destBuf, word);
                    }

                    break;

                case mode_no_compress:
                default:

                    break;
                }

                continue;
            }
	    }
	
	    if(q) 
	    {
		    q--; 
		    
		    if(q > 0x1F) 
			    PutByte( destBuf, 0xE0 + (q >> 8) );
		    
		    PutByte(destBuf, q);
		    q++;
	    
            CopyBuffer(destBuf, srcBuf,
                       destBuf->wOffset, i - q,
                       q);

            destBuf->wOffset += q; 
	    }
	    
        // ensure that the data we currently have is terminated and won't
        // crash the game
	    PutByte(destBuf, 0xFF);

        // eliminates excess allocated space and resizes the buffer
        // to just fit the data we have
        ResizeBuffer(destBuf, destBuf->wOffset);
	    
	    return destBuf;
    }

// ===============================================================

    void Do3To4Low(bufPtr targBuf, bufPtr srcBuf, uint32_t *targPtr, uint32_t *srcPtr, uint32_t tiles)
    {
        // Upconverts 3bpp graphics to 4bpp, and restricts to the lower 8 colors of the palette
        // In other words it ensures that the 4th byte of each line in a tile is always 0x00

        uint32_t i = 0;
        uint32_t j = 0;

        // ------------------------------

        for(i = 0; i < tiles; i++)
        {
            CopyBufferWB(targBuf, srcBuf,
                         targPtr, srcPtr,
                         0x10);
    
            for(j = 0; j < 16; j += 2)
            {
                Put2Bytes(targBuf, *targPtr, GetByte(srcBuf, *srcPtr));
                *targPtr += 2;
                (*srcPtr)++;
            }
        }
    }

// ===============================================================

    void Do3To4High(bufPtr targBuf, bufPtr srcBuf, uint32_t *targPtr, uint32_t *srcPtr, uint32_t cap)
    {
        // upconverts 3bpp graphics to 4bpp and makes is so that any nonzero color index
        // has the high bits set (meaning it uses the upper 8 colors of the palette

        uint32_t oldPos = 0;
        uint32_t start = (*srcPtr);
        uint32_t i = start;
        uint32_t j = (*targPtr);
        uint32_t k = 0;

        u8 value = 0, accum = 0;

        // ------------------------------

        for( ; i < srcBuf->length; )
        {
            if((i - start) >= cap) 
                return;

            // save this position - we'll need it.
            oldPos = i;

            CopyBufferWB(targBuf, srcBuf,
                         &j,      &i,
                         0x10);

            for(k = 0; k < 0x10; i++, j += 2, k += 2)
            {
                value = GetByte(srcBuf, i);

                accum = value;
                accum |= GetByte(srcBuf, oldPos + k);
                accum |= GetByte(srcBuf, oldPos + k + 1);

                PutByte(targBuf, j, value);
                PutByte(targBuf, j + 1, accum);
            }
        }

        (*srcPtr) = i;
        (*targPtr) = j;
    }

// ===============================================================

    void LoadTilesets(zgPtr game)
    {
        // this routine corresponds to $00:E19B in the rom

        uint32_t i = 0, j = 0;
        uint32_t index = 0, index2 = 0, index3 = 0, index4 = 0, dummyInt = 0, source = 0; 
        
        bufPtr decomp = NULL;
        giPtr gi = game->gi;
        
        // -----------------------------------------------
        
        // Find the pointer to the graphics to decompress
        // should optionally use this or 0x01 or 0x0A depending on whether you're in the dark world or not.
        index4 = game->GetSprGfxPtr(gi->spriteGfx1);
        
        // Decompress them using a special method
        index = 0x8800;
        
        Do3To4High(game->vram, game->image,
                   &index,     &index4,
                   0x600);
        
        // --------------------------------------------------------
        
        // 0x06 is always used for this graphics slot
        index4 = game->GetSprGfxPtr(0x6);
        
        Do3To4Low(game->vram, game->image,
                  &index,     &index4,
                  0x80);
 
        // --------------------------------------------------------

        // Now we move on to decompressing and storing the 4 graphics subsets
        // that are dependent upon the room you are in.

        index3 = gi->spriteGfx0;
        index3 *= 4;
        index3 += 0x5B57;

        for(i = 0; i < 4; i++, index3++)
        {
            // Get one of the current 4 graphics set numbers
            index2 = GetByte(game->image, index3);
            index4 = game->GetSprGfxPtr(index2);
    
            switch(index2)
            {
            case 0x00:

                for(j = 0; j < 0x800; j++, index++)
                    PutByte(game->vram, index, 0);

                break;

            case 0x52:
            case 0x53:
            case 0x5A:
            case 0x5B:
            case 0x5C:
            case 0x5E:
            case 0x5F:

                decomp = DecompressStandard(game->image, index4);
                index4 = 0;

                Do3To4High(game->vram, decomp,
                           &index,     &index4,
                           0x600);

                DeallocBuffer(decomp);

                break;

            default:
            
                decomp = DecompressStandard(game->image, index4);
                index4 = 0;

                Do3To4Low(game->vram,   decomp,
                          &index,       &index4,
                          0x40);

                DeallocBuffer(decomp);

                break;
            }
        }



        // Next decompress the BG graphic data
        // Target in VRAM is 0x4000

        index = 0x4000;

        // The primary BG tile set 
        index3 = gi->backGfx0;
        index3 *= 8;
        index3 += 0x6073;

        // graphicsNum corresponds to $0AA2
        index2 = gi->backGfx1;
        index2 *= 4;
        index2 += 0x5D97;

        for(i = 0; i < 8; i++, index3++)
        {
            j = 7 - i;

            source = GetByte(game->image, index3);     
            source = game->GetBgGfxPtr(source);

            index4 = GetByte(game->image, index2 + i - 3);

            if( j >= 1 && j <= 4)
                if( index4 != 0)
                    source = game->GetBgGfxPtr(index4);

            index4 = 0;

            decomp = DecompressStandard(game->image, source);

            // depending on the value of $0AA1 we use different 3bpp to 4bpp methods
            if( (gi->backGfx0 < 0x20) && (j < 3))
                Do3To4Low(game->vram, decomp, &index, &index4, 0x40);
            else if(gi->backGfx0 < 0x20)
                Do3To4High(game->vram, decomp, &index, &index4, 0x600);
            else 
            {
                switch(j)
                {
                case 2:
                case 3:
                case 4:
                case 7:
                    Do3To4High(game->vram, decomp, &index, &index4, 0x600);
                    break;

                default:
                    Do3To4Low(game->vram, decomp, &index, &index4, 0x40);
                    break;
                }
            }

            DeallocBuffer(decomp);
        
        }

        return;
    }

// ===============================================================

    void LoadPalettes(zgPtr game, uint32_t offset, uint32_t numPalettes, uint32_t numColors, uint32_t cgOffset)
    {
        bufPtr rom   = game->image;

        uint32_t i        = 0;
        uint32_t ret      = 0;
        uint32_t numBytes = (numColors * 2);

        // ------------------------------

        for(i = 0; i < numPalettes; i++, offset += numBytes)
        {
            ret = CopyBuffer(game->cgram,         rom,
                             cgOffset + (i*0x20), offset,
                             numBytes);
        }
    }

// ===============================================================

    void LoadPalettes(zgPtr game)
    {
        // Notes on terminology

        // BP = Background Palette
        // SP = Sprite Palette
        // e.g. BP-0 would refer to the first Background palette
        // which is $0-$1F in CGRAM
        // SP-0 would be the first sprite palette ($100-$11F)
        // All writes to CGRAM are indented by 2 bytes
        // Partially because of the way LTTP was designed to swap 3bpp graphics from 
        // "side to side," so to speak,
        // and partically because the first color of every 16 color palette is ignored.
        // (or 4 or 256 color palette for that matter)

        //   Halves of each palette     Memory Involved
        //   first  | second
        /*
        BP-0 $DEE52 | (full)            $0AB2
        BP-1 $DEE52 | ( .. )            ( .. )
        BP-2 $DEE74 | ( .. )            $0AB6
        BP-3 $DEE74 | ( .. )            ( .. )
        BP-4 $DEE74 | ( .. )            ( .. )
        BP-5 $DEE74 | ( .. )            ( .. )
        BP-6 $DEE74 | ( .. )            ( .. )
        BP-7 $DEE74 | ( .. )            ( .. )

        SP-0 $DEC77 | $DEE74            $0AAC | $0AB6
        SP-1 $DEC9E | (full)            lightworld / darkworld
        SP-2 $DEC9E | ( .. )            ( .. )                        
        SP-3 $DEC9E | ( .. )            ( .. )
        SP-4 $DEC9E | ( .. )            ( .. )
        SP-5 $DECC5 | $DED03 & $DED29   $0AAD | $7EF359-A
        SP-6 $DECE4 | $DED6E            $0AAE | $0AB1 (different for outdoors)
        SP-7 $DEDF9 | (full)            $
        */

        uint32_t bp_2bpp = 0; // could be 1 in some situations like ending mode.
        uint32_t paletteNum = game->bm_dung_header.paletteIndex * 4;

        uint32_t bp_4bpp = GetByte(game->image, paletteNum + 0x75460);
        uint32_t sp_0 = GetByte(game->image, paletteNum + 0x75461);
        uint32_t sp_1 = GetByte(game->image, paletteNum + 0x75462);
        uint32_t sp_2 = GetByte(game->image, paletteNum + 0x75463);
        uint32_t sword = 4, shield = 3, armor = 2;
        uint32_t sp_throw = 0;

        uint32_t source;
        uint32_t bank = 0x1B0000;
        uint32_t i = 0;

        BM_DungeonHeader *header = &game->bm_dung_header;
        bufPtr rom = game->image;
        bufPtr cgram = game->cgram;

        const uint32_t full = 0x1E, second = 7*2;

        // ------------------------------

        if(game->currentFunc == dung)
            sp_throw = 0xA;

        // Loads BP-0 through BP-1 (full) --------------------------
        source = GetByte(rom, bp_2bpp + 0xDEC47);
        source = CpuToRomAddr( (source + 0xD660) | bank ) ;

        CopyBuffer(cgram, rom, 0, source, 0x20 * 2);
    
        // Loads BP-2 through BP-7 and SP-0 (second half) ----------
        source = Get2Bytes(rom, bp_4bpp + 0xDEC4B);
        source = CpuToRomAddr( (source + 0xD734) | bank );
    
        CopyBuffer(cgram, rom, 0x112, source, second); 

        for(i = 0; i < 6; i++, source += full)
        {
            CopyBuffer(cgram,               rom,
                       0x42 + (i * 0x20),   source,
                       full);
        }

        // Loads SP-0 (first half) ---------------------------------
        source = GetByte(rom, sp_0 + 0xDEBC6);
        source = CpuToRomAddr( (source + 0xD39E) | bank );

        CopyBuffer(cgram, rom, 0x102, source, second);

        // Loads SP-1 through SP-4 (full)
        // Note: look up later how to find the world for each room
        source = game->currentDungRoom & 0x40 ? 2 : 0;
        source = 0;
        source = Get2Bytes(rom, source + 0xDEC73);
        source = CpuToRomAddr( (source + 0xD218) | bank );

        for(i = 0; i < 4; i++, source += full)
        {
            CopyBuffer(cgram,                   rom,
                       0x122 + (i * 0x20),      source,
                       full);
        }

        // Loads SP-5 (first half)
        
        source = Get2Bytes(rom, (sp_1 * 2) + 0xDEBD6);
        source = CpuToRomAddr( (source + 0xD4E0) | bank );
        
        CopyBuffer(cgram, rom, 0x1A2, source, second);
        
        // Loads SP-5 (second half)
        // Uses a sword value to load a 3 color palette
        
        source = GetByte(rom, sword + 0xDEBB4);
        source = CpuToRomAddr( (source + 0xD630) | bank );
        
        CopyBuffer(cgram, rom, 0x1B2, source, 6);
        
        // Uses the shield value to load a 4 color palette
        source = GetByte(rom, shield + 0xDEBC1);
        source = CpuToRomAddr( (source + 0xD648) | bank );
        
        CopyBuffer(cgram, rom, 0x1B8, source, 8);
        
        // Loads SP-6 (first half)
        
        source = Get2Bytes(rom, (sp_2 * 2) + 0xDEBD6);
        source = CpuToRomAddr( (source + 0xD4E0) | bank );
        
        CopyBuffer(cgram, rom, 0x1C2, source, second);
        
        // Loads SP-6 (second half)
        
        source = GetByte(rom, sp_throw + 0xDEBC6);
        source = CpuToRomAddr( (source + 0xD446) | bank );
        
        CopyBuffer(cgram, rom, 0x1D2, source, second);
        
        // Loads SP-7 (full)
        
        source = GetByte(rom, armor + 0xDEC06) << 1;
        source = CpuToRomAddr( (source + 0xD308) | bank );
        
        CopyBuffer(cgram, rom, 0x1E2, source, full);
    }

// ===============================================================

    SnesTile* SnesToTile(bufPtr image, SnesTile* product, unsigned int offset)
    {
        // see planning.c for more info on how this works.

        u8 color_bits[4];
        u8 color_value = 0;

        int line = 0;
        int column = 0;
        int masked_bits;
        int currentline = 0;

        uint32_t mask;

        // ------------------------------

        for(  ; line < 8; line++)
        {
            currentline = Get2Bytes(image, (line * 2) + offset);
            currentline |= ( Get2Bytes( image, (line * 2) + 0x10 + offset) << 0x10 ); 

            mask = 0x80808080;

            for( column = 0; column < 8; column++)
            {
                masked_bits = (currentline & mask);
        
                color_bits[0] = (u8) (masked_bits);
                color_bits[0] >>= (7 - column);
        
                color_bits[1] = (u8) (masked_bits >> 0x08);
                color_bits[1] >>= (7 - column);
                color_bits[1] <<= 1;
        
                color_bits[2] = (u8) (masked_bits >> 0x10);
                color_bits[2] >>= (7 - column); 
                color_bits[2] <<= 2;

                color_bits[3] = (u8) (masked_bits >> 0x18);
                color_bits[3] >>= (7 - column);
                color_bits[3] <<= 3;

                color_value = color_bits[0] | 
                              color_bits[1] | 
                              color_bits[2] | 
                              color_bits[3];

                color_value &= 0x0F;
        
                (*product)[column + (8*(7-line))] = color_value;

                mask >>= 1;
            }
        }

        return product;
    }

// ===============================================================

    SnesTile* FlipTile(SnesTile* input, SnesTile* product, int x, int y)
    {
        int column = 0;
        int line = 0;

        SnesTile temp;

        // ------------------------------

        if(!x && !y)
        {
            memcpy(*product, *input, sizeof(SnesTile));

            return product;
        }

        // if we need to x flip, do so.
        if(x)
        {
            column = 0;
            line = 0;

            while(line < 64)
            {
                for( column = 0; column < 8; column++)
                {
                    (*product)[column + line] = (*input)[(7 - column) + line];
                }

                line += 8;
            }

            memcpy(temp, *product, sizeof(SnesTile));
        }
        else
            memcpy(temp, *input, sizeof(SnesTile));



        // same with y flip.
        if(y)
        {
            line = 0;
            column = 0;

            while( column < 8)
            {
                for( line = 0; line < 64; line += 8)
                {
                    (*product)[line + column] = (temp)[(56 - line) + column];
                }

                column++;
            }

        }

        return product;
    }

// ===============================================================

    int TileToBitmap(SnesTile* tile, SnesImage* product, bufPtr cgram, uint32_t palette)
    {
        /*
        int index = 0;
        unsigned char color_value;

        for( ; index < 64; index++)
        {
            color_value = (*tile)[index];

            (*product)[index] = (*palette)[color_value];
        }
        */
        return 1;
    }

// ===============================================================

    GameBoyImage* GameBoyToBitmap(bufPtr image, GameBoyPal* palette, GameBoyImage* product)
    {

        return product;
    }

// ===============================================================

    ManagedBitmap* CreateManagedBitmap(uint32_t width, uint32_t height)
    {
        uint32_t headerSize = sizeof(BITMAPINFOHEADER);    
        uint32_t numPixels = width * height;
    
        ManagedBitmap* b = (ManagedBitmap*) malloc(sizeof(ManagedBitmap));
    
        // --------------------------------------

        if(!b)
            return NULL;

        b->memory = CreateBuffer(headerSize + (sizeof(RGBQUAD) * numPixels) );
        b->header = (BITMAPINFOHEADER*) (b->memory->contents);
        b->structure = (BITMAPINFO*) (b->memory->contents);

        b->header->biBitCount = 32;
        b->header->biClrImportant = 0;
        b->header->biClrUsed = 0;
        b->header->biCompression = BI_RGB;
        b->header->biHeight = height;
        b->header->biPlanes = 1;
        b->header->biSize = headerSize;
        b->header->biSizeImage = 0;
        b->header->biWidth = width;
        b->header->biXPelsPerMeter = 0;
        b->header->biYPelsPerMeter = 0;

        b->structure->bmiHeader = *(b->header);

        b->pixels = GetSubBuffer(b->memory, headerSize);
        b->stride = (b->header->biWidth) * 4;
        b->bottom = b->stride * (b->header->biHeight - 1);

        return b;
    }


// ===============================================================

    void DecodePalettes(zgPtr game)
    {
        uint32_t color   = 0;
        uint32_t i       = 0;
        uint32_t j       = 0;

        // masks for extracting each 5-bit component
        // of color data from the 15-bit BGR (SNES) color.
        uint32_t red     = 0x1F;
        uint32_t green   = 0x1F << 5;
        uint32_t blue    = 0x1F << 10;

        bufPtr cgram = game->cgram;

        // --------------------------
    
        for(i = 0; i < 0x10; ++i)
        {
            for(j = 0; j < 0x10; ++j)
            {
                color = Get2Bytes(cgram, (0x20*i) + (2*j));
            
                (game->Pals[i])[j]  = ((color & red)   << 0x13);
                (game->Pals[i])[j] |= ((color & green) << 0x06);
                (game->Pals[i])[j] |= ((color & blue)  >> 0x07);
                (game->Pals[i])[j] &= 0xFFFFFF;

            }
        }
    }

// ===============================================================

    // The default offset is 0. In the case of animated tiles you might
    // use a later offset to only exclude
    void DecodeTiles(zgPtr game, uint32_t firstTile, uint32_t lastTile)
    {
        bufPtr vram = game->vram;

        uint32_t i = firstTile;
        uint32_t j = 0x20 * firstTile;

        for( ; i <= lastTile ; ++i, j += 0x20)
            SnesToTile(vram, &game->BGTiles[i], 0x4000 + j);

        return;
    }

// ===============================================================

    void DecodeOAMTiles(zgPtr game)
    {
        bufPtr vram = game->vram;

        uint32_t i = 0;
        uint32_t j = 0;

        for(i = 0; i < 0x200; ++i, j += 0x20)
            SnesToTile(vram, &game->OAMTiles[i], 0x8000 + j);
    }

// ===============================================================
