
#include "overdata.h"

namespace Overworld
{
    const uint32_t mask16      = ~0x01;
    const uint32_t mask32      = ~0x03;
    
// ===============================================================

    OverData::OverData(Buffer *rom)
    {
        // copy of the rom from the ZeldaGame object
        this->rom           = rom;
        
        // Initializes all the map16 and map32 flags so we can know which tiles are being used and which aren't.
        LoadMapFlags();
        
        // load various types of different data
        LoadOverlaysOld();
        LoadAllSpriteData();
        LoadAllItemData();
    }

// ===============================================================

    OverData::~OverData()
    {
        
    }

    static FILE* mapCountFile = fopen("C:\\map32log.log", "wt");

// ===============================================================

    Map16Val const OverData::reservedMap16[] = 
    {
        stock_empty,
        stock_bush,
        stock_bush2,
        stock_rock_sl,
        stock_rock_sh,
        stock_peg,
        stock_rock_bl0,
        stock_rock_bl1,
        stock_rock_bl2,
        stock_rock_bl3,
        stock_rock_bh0,
        stock_rock_bh1,
        stock_rock_bh2,
        stock_rock_bh3,
        stock_warp,
        stock_rockPile0,
        stock_rockPile1,
        stock_rockPile2,
        stock_rockPile3,
        stock_sign,
        stock_grass,
        stock_jumpPoint,
        stock_stump0,
        stock_stump1,
        stock_stump2,
        stock_stump3,
        0x0212,
        0x0DA4,
        0x0DA5,
        0x0DA6,
        0x0DA7,
        0x0DB5,
        0x0DC7,
        0x0DC8,
        0x0E1B,
        0x0E21,
        0x0E25,
        0x0E3F,
        0x0E48,
        0x0E54,
        0x0E64,
        0x0E88,
        0x0E8C,
        0x0E92,
        0x0E9A,
        0x0E9C
    };    

// ===============================================================

    bool OverData::LoadArea()
    {
        bool result     = true;
        bool darkWorld  = false;
        
        // ----------------------------
        
        // refresh the tiles in the map8 buffer
        LoadMap8();
        
        return true;
    }

// ===============================================================

    uint32_t OverData::FindMap16(uint16_t map8Vals[4])
    {
    #if(0)
        uint16_t i = 0;
        
        const uint16_t theEnd = 0x0DF4;
        
        for(i = 0; i < 0x3000; i++)
        {
            if(i == theEnd) // Eureka Seven reference
                return -1;
            
            if(map16To8[i].upperLeft       != map8Vals[0])
                continue;
            else if(map16To8[i].upperRight != map8Vals[1])
                continue;
            else if(map16To8[i].lowerLeft  != map8Vals[2])
                continue;
            else if(map16To8[i].lowerLeft  != map8Vals[3])
                continue;
            else
                return (((uint32_t) i) & 0x0000FFFF);
        }
        
    #endif
        return -1;
    }

// ===============================================================

    uint32_t OverData::AllocateMap16(uint16_t map8Vals[4], uint16_t oldMap16, uint16_t threshold)
    {
    #if(0)
        uint32_t a = 0;
        uint32_t i = 0;
        
        for(i = threshold; i < 0x4000; i++)
        {
            // Get the number of times this particular map16 value occurs in the rom.
            a = map16Counts->Get4Bytes(i * 4);
            
            if(a > 1)
                continue;
            else if(a == 1)
            {
                // in this case the new map16 tile would replace the old
                if( (uint16_t) i == oldMap16 )
                    return i;
                
                continue;
            }
            else
                return i;
        }
        
    #endif
        return -1;
    }

// ===============================================================

    void OverData::DecMapCounts(uint16_t mapVal)
    {
    #if(0)
        uint16_t     *total  = &(this->numMap16Tiles);
        
        uint32_t     a       = 0;
        
        bufPtr  counts  = this->map16Counts;
        bufPtr  flags   = this->map16Flags;
        
        // ------------------------------------------------------
        
        a = counts->Get4Bytes(mapVal * 4);
        
        // subtract one but don't wrap under 0
        if(a)
            a = a - 1;
        
        if(!a)
            a = a;
        
        counts->Put4Bytes(mapVal * 4, a);
        
        if(a)
            flags->PutBit(1, mapVal);
        else
        {
            flags->PutBit(0, mapVal);
            (*total) -= 1;
        }
    #endif
    }

// ===============================================================

    void OverData::IncMapCounts(uint16_t mapVal, uint16_t *map2x2)
    {
    #if(0)
        uint16_t     *total  = &(this->numMap16Tiles);

        uint32_t     a       = 0;
        uint32_t     offset  = mapVal * 8;

        bufPtr  counts  = this->map16Counts;
        bufPtr  flags   = this->map16Flags;
        
        // ------------------------------------------------------
        
        a = counts->Get4Bytes(mapVal * 4);
        
        if(a == 0xFFFFFFFF)
            a = a;
        
        if(!a)
            a = a;
        
        a = (a < 0xFFFFFFFF) ? a + 1 : a; // add one but don't wrap past 0xFFFFFFFF
        
        counts->Put4Bytes(mapVal * 4, a);
        flags->PutBit(1, mapVal);
        
        // On a zero to one transition on counts, we replace
        // the map16 to map8 correspondances, as they might have changed.
        if( (a == 1) && (map2x2) )
        {
            map16To8[mapVal].upperLeft  = map2x2[0];
            map16To8[mapVal].upperRight = map2x2[1];
            map16To8[mapVal].lowerLeft  = map2x2[2];
            map16To8[mapVal].lowerRight = map2x2[3];
            
            (*total) += 1;
        }
    #endif
    }

// ===============================================================

    bufPtr OverData::LoadMap8()
    {
    #if(0)
        uint16_t map8Vals[4]; 

        uint32_t i = 0;
        uint32_t j = 0;
        uint32_t k = 0;
        uint32_t index = 0;
        
        // --------------------------------
        
        // convert the map16 data into map8 tiles
        for
        (
            i = 0, j = 0, k = 0;
            i < map16Buf->length;
            i += 2, j += 4, k += 4
        )
        {
            if(k == 0x100)
            {
                k = 0;
                j += 0x100;
            }
            
            index = Get2Bytes(map16Buf, i);
            
            map8Vals[0] = map16To8[index].upperLeft;
            map8Buf->Put2Bytes(j, map8Vals[0]);
            
            map8Vals[1] = map16To8[index].upperRight;
            map8Buf->Put2Bytes(j + 0x02, map8Vals[1]);
            
            map8Vals[2] = map16To8[index].lowerLeft;
            map8Buf->Put2Bytes(j + 0x100, map8Vals[2]);
            
            map8Vals[3] = map16To8[index].lowerRight;
            map8Buf->Put2Bytes(j + 0x102, map8Vals[3]);
        }
        return map8Buf;
    #endif
    
        /// hurp durp
        return new Buffer(42);
    }

// ===============================================================

    void OverData::LoadMapFlags()
    {
    #if(0)
        uint16_t map16Val = 0;
        
        uint32_t i = 0, j = 0;
        uint32_t numReserved = sizeof(reservedMap16) / sizeof(uint16_t);
        
        bufPtr map16Buf = this->map16Buf;
        
        // -----------------------------------
        
        for(i = 0; i < 0xA0; i += 4)
        {
            LoadMap16(i + 0, upper_left);
            LoadMap16(i + 1, upper_right);
            LoadMap16(i + 2, lower_left);
            LoadMap16(i + 3, lower_right);

            // At this point we're done loading the 4 map32 sets and now we process the map16
            // Flags for the result of that.
            for(j = 0; j < map16Buf->length; j += 2)
            {
                map16Val = Get2Bytes(this->map16Buf, j);
                IncMapCounts(map16Val);
            }
        }
        
        // Make sure to protect the map16 tiles that have special purposes
        for(i = 0; i < numReserved; ++i)
            IncMapCounts(reservedMap16[i]);

        // counting the number of used map16 tiles
        for(j = 0, i = 0; i < this->map16Flags->length * 8; i++)
        {
            if(GetBit(this->map16Flags, i))
                j++;
        }

        this->numMap16Tiles = j;
        
    #endif
    }

// ===============================================================

    bool OverData::LoadAllSpriteData()
    {
    #if(0)
        bool overlord       = false;
        bool usingOld       = false;

        // sprite id, x coordinate, and y coordinate
        uint8_t  id      = 0, x      = 0, y = 0;

        uint32_t phase   = 0, area   = 0;
        uint32_t i       = 0, j      = 0, k = 0;
        uint32_t value   = 0, offset = 0;

        uint32_t tableSizes[3]   = { 0x90, 0x90, 0x90 }; // number of areas in each table
        uint32_t tableOffset[3]  = { header->overSprOffset0, header->overSprOffset2, header->overSprOffset3 };

        OverSpr *s      = NULL;
        OverSpr **list  = NULL;

        OverArea *a     = NULL;
        
        // ---------------------------
        
        usingOld = (tableOffset[0] == 0x9C881) ? true : false;

        // after the rom has been saved once by Black Magic, sprites are stored for all 0xC0 areas, rather than the first 0x90.
        if(!usingOld) 
        { 
            for(i = 0; i < 3; ++i) 
            {
                tableSizes[3] = 0xC0;
            }
        }

        for(phase = 0; phase < 3; ++i)
        {
            for(area = 0; area < tableSizes[phase]; ++area)
            {
                // only going to operate on canonical area numbers
                // to avoid duplication of data loading
                if(ResolveArea(area) != area)
                    continue;

                a                       = areas[area];

                list                    = &a->spr[phase];

                a->fallingRocks[phase]  = false;

                a->spr[phase]           = NULL;

                // dark world and extended areas in the "beginning" part don't have sprites in the original rom.
                if( (usingOld) && (phase == 0) && (area >= 0x40) )
                    continue;

                offset = CpuToRomAddr(tableOffset[phase]);

                // the new mechanism and the old one are different in that the old mechanism uses 2-byte pointers in the same bank as the table.
                offset = (usingOld) ? CpuToRomAddr( (0x090000) | Get2Bytes(rom, offset + (area * 2) ) ) : 
                                      CpuToRomAddr( Get3Bytes(rom, offset + (area * 3) ) );
               
                for(i = offset; (i & 0xFFFF) < 0xFFFD; i += 3)
                {
                    y   = GetByte(rom, i);

                    if(y == 0xFF)
                       break;

                    x   = GetByte(rom, i + 1);
                    id  = GetByte(rom, i + 2);

                    // falling rocks don't really count as a sprite so we skip it
                    if(id == 0xF4)
                    {
                        a->fallingRocks[phase] = true;
                        continue;
                    }
                    else if(id >= 0xF5)
                    {
                        // overworld overlords are not actually implemented yet
                        overlord = true;
                        id -= 0xF5;
                    }

                    OverSpr::Add(list, new OverSpr(id, x << 4, y << 4) ); 
                }

            }

            for(area = tableSizes[i]; area < 0xC0; ++area)
            {
                if(ResolveArea(area) != area)
                    continue;

                areas[area]->spr[phase] = NULL;
            }
        }
    #endif
	    return true;
    }

// ===============================================================

    bool OverData::LoadAllItemData()
    {
    #if(0)
        uint32_t i       = 0;
        uint32_t area    = 0;
        uint32_t x       = 0, y = 0;
        uint32_t map16   = 0;
        uint32_t offset  = 0;
        uint32_t itemNum = 0;

        OverArea *a = NULL;

        // ---------------------------
        
        for(area = 0; area < 0xC0; ++area)
        {
            // There's no item data in the extended areas
            // Though we might put some there eventually
            if(area >= 0x80)
                continue;

            // don't operate on unauthentic area numbers.
            if(ResolveArea(area) != area)
                continue;

            a = areas[area];

            if(header->overItemOffset == 0x1BC2F9)
            {
                // older method
                offset = CpuToRomAddr(header->overItemOffset + (i * 2) );
                offset = CpuToRomAddr(Get2Bytes(rom, offset) | (0x1B << 16) );
            }
            else
            {
                // newer method
                offset = CpuToRomAddr(header->overItemOffset + (i * 3) );
                offset = CpuToRomAddr(Get3Bytes(rom, offset));
            }
    
            for(i = 0;  ; i += 3)
            {
                map16  = Get2Bytes(rom, offset + i);

                if(map16 == 0xFFFF)
                    break;
                
                x = GetMap16X(map16);
                y = GetMap16Y(map16);

                itemNum = GetByte(rom, offset + i + 2);
                
                if(itemNum > 0x16)
                {
                    switch(itemNum)
                    {
                        case 0x80:
                        case 0x82:
                        case 0x84:
                        case 0x86:
                        case 0x88:

                            break; // valid item numbers

                        default:

                            itemNum = 0x00;
                   }
                }

                // In the original game, changing the items available in each phase
                // I'm toying with the idea of doing that, though.
                // For now all items reside in phase 0.
                OverItem::Add( &a->items[0], new OverItem(itemNum, x, y));
            }
        }
    #endif
        return true;
    }

// ===============================================================

    void OverData::LoadEventOverlay(uint16_t areaNum, Map16Buf &lay)
    {
        bool done  = false;
        bool write = false;
        
        uint8_t opcode = 0;
        
        POINT min = {0,    0};
        POINT max = {0x3F, 0x3F};
        
        uint32_t map16Val  = 0;
        uint32_t map16Addr = 0;
        uint32_t offset    = 0;
        uint32_t x_reg     = 0;
        
        // -----------------------
        
        done   = false;
        
        // Get a pointer to the overlay's routine in the rom.
        offset = m_rom->Get2Bytes(0x77664 + (areaNum << 1) ) + (0x0E << 0x10);
        offset = CpuToRomAddr(offset);
        
        // handles parsing of different ASM opcodes because 
        // the overlays are all stored as ASM instructions rather than data... (Yeah, dumb, I know)
        
        while(!done)
        {
            write  = false;
            opcode = m_rom->GetByte(offset++);
            
            switch(opcode) 
            {
                // INC A instruction (increments the map16 value.)
                case 0x1a:
                    map16Val += 1;
                    break;
                
                // JMP instruction (tells us to go to another location in the rom to finish the routine)
                case 0x4c:
                    offset = m_rom->Get2Bytes(offset) + (0x0E << 0x10);
                    offset = CpuToRomAddr(offset);
                    break;
                
                // RTS instruction (ends the routine)
                case 0x60:
                    done = true;
                    break;
                
                // STA absolute instruction. The map16 address is the argument.
                case 0x8d:
                    map16Addr = m_rom->Get2Bytes(offset);
                    offset += 2;
                    write = true;
                    break;
                
                // STA long instruction. Uses a 2-byte indexed pointer
                case 0x9d:
                    // Get the 3-byte pointer but strip off the 0x7E.
                    map16Addr = m_rom->Get2Bytes(offset) + x_reg;
                    
                    offset += 2;
                    write = true;
                    break;
                
                // STA long instruction.
                case 0x8f:
                    // Get the 3-byte pointer but strip off the 0x7E.
                    map16Addr = m_rom->Get3Bytes(offset);
                    
                    if(GetBank(map16Addr) != 0x7E)
                        done = true;
                    else
                        map16Addr &= 0xFFFF;
                    
                    offset += 3;
                    write = true;
                    break;
                
                // STA long indexed instruction.
                case 0x9f:
                    map16Addr = m_rom->Get3Bytes(offset) + x_reg;
                    
                    // bad bank, must be 0x7E!
                    if(GetBank(map16Addr) != 0x7E)
                        done = true;
                    else
                        map16Addr &= 0xFFFF;
                    
                    offset += 3;
                    write = true;
                    break;
                
                case 0xa2:
                    x_reg = m_rom->Get2Bytes(offset);
                    offset += 2;
                    break;
                
                case 0xa9:
                    map16Val = m_rom->et2Bytes(offset);
                    offset += 2;
                    break;
                
                case 0xea:
                    offset = offset;
                    break;
                
                default:
                    done = true;
                    break;
            }
            
            if(offset > 0x78000)
                done = true;
            
            if(write)
            {
                if((map16Addr < 0x2000) || (map16Addr > 0x3FFF))
                    done = true;
                else
                {
                    SetTile(map16Addr - 0x2000, map16Val))
                    
                    /// IncMapCounts(map16Val, false);
                }
            }
        }
    }

// ===============================================================
}