
    #include <deque>
    #include <list>
    
    #include "inttypes.h"
    
    #include "containers.h"
    
#ifndef OVERTYPES_H

    #define OVERTYPES_H

namespace Overworld
{
    using std::list;
    using std::exception;
    
    using namespace Containers;
    using namespace Graphics;
    
    template<uint16_t grid = 1>
    class Position
    {
        
    protected:
        
        uint16_t m_x;
        uint16_t m_y;
        
    public:
        
        Position(uint16_t p_x = 0,
                 uint16_t p_y = 0)
        {
            Set(p_x, p_y);
        }
        
        Position(Position const &p)
        {
            Set(p.m_x, p.m_y);
        }
        
        Position& operator = (Position const &p)
        {
            Set(p.m_x, p.m_y);
        }
        
        void Set(uint16_t p_x,
                 uint16_t p_y)
        {
            m_x = ( p_x - (p_x % grid) );
            m_y = ( p_y - (p_y % grid) );
        }
        
        uint16_t GetX()
        {
            return m_x;
        }
        
        uint16_t GetY()
        {
            return m_y;
        }
    };
    
// ================================================
    
    class Entrance
    {
        
    public:
        
        uint16_t m_entrance;
        
        Position<16> m_pos;
        
    public:
        
        Entrance(uint16_t x = 0, uint16_t y = 0, uint16_t entrance = 0)
        : m_pos(x, y),
          m_entrance(entrance)
        {
            
        }
    };
    
    typedef list<Entrance> EntranceList;


// ================================================
    
    class Exit
    {
        
    public:
        
        enum exit_type
        {
            down,
            up,
            wooden,
            bombable,
            sanctuary,
            palace
        };
        
        exit_type m_type;
        
        uint16_t m_area;
        
        uint16_t m_room;
        
        uint16_t m_doorType;
        
        Position<1> m_pos;
        Position<1> m_doorPos;
        
    public:
        
        Exit(uint16_t x     = 0,
             uint16_t y     = 0,
             uint8_t area   = 0,
             uint16_t room  = 0,
             exit_type type = up)
        
        : m_pos(x, y),
          
          m_area(area),
          m_room(room),
          m_type(type)
        {
            
        }
        
        Exit(Exit const &p_e)
        
        : m_type(p_e.m_type),
          
          m_area(p_e.m_area),
          m_room(p_e.m_room),
          m_doorType(p_e.m_doorType),
          
          m_pos(p_e.m_pos),
          m_doorPos(p_e.m_doorPos)
        {
        }
        
        Exit & operator = (Exit const &p_e)
        {
            m_type = p_e.m_type;
            
            m_area     = p_e.m_area;
            m_room     = p_e.m_room;
            m_doorType = p_e.m_doorType;
            
            m_pos     = p_e.m_pos;
            m_doorPos = p_e.m_doorPos;
        }
    };
    
    typedef list<Exit> ExitList;

// ================================================
    
    class Sprite
    {
        
    public:
        
        bool overlord;
        
        uint8_t exData1;
        uint8_t exData2;
        uint8_t exData3;
        
        uint16_t spriteNum;
        
        Position<16> pos;
        
    public:
        
        Sprite(uint16_t _spriteNum = 0,
               uint16_t x = 0,
               uint16_t y = 0,
               bool _overlord = false,
               uint8_t _exData1 = 0,
               uint8_t _exData2 = 0,
               uint8_t _exData3 = 0)
        
        : overlord(_overlord),
          spriteNum(_spriteNum),
          pos(x, y),
          exData1(_exData1),
          exData2(_exData2),
          exData3(_exData3)
        {
            SetType(spriteNum);
        }
        
    // >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        
        bool SetType(uint16_t value)
        {
            bool retVal = false;
            
            if(value <= 0x00F3)
            {
                spriteNum = value;
                retVal    = true;
            }

            return retVal;
        }
    };
    
    typedef list<Sprite> SpriteList;
    
// ================================================
    
    class Item
    {
        
    public:
        
        uint8_t m_exData1;
        uint8_t m_exData2;
        uint8_t m_exData3;
        
        uint16_t m_itemNum;
        
        Position<16> m_pos;
        
    public:
        
        Item(uint16_t p_itemNum = 0,
             uint16_t p_x = 0,
             uint16_t p_y = 0,
             uint8_t  p_exData1 = 0,
             uint8_t  p_exData2 = 0,
             uint8_t  p_exData3 = 0)
             
        : m_pos(p_x, p_y),
          m_exData1(p_exData1),
          m_exData2(p_exData2),
          m_exData3(p_exData3)
        {
            SetType(p_itemNum);
        }
        
    // >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        
        bool SetType(uint16_t value)
        {
            bool retVal = false;
            
            if(value < 16)
            {
                m_itemNum = value;
                
                retVal = true;
            }
            
            switch(value)
            {
                case 0x80:
                case 0x82:
                case 0x84:
                case 0x86:
                case 0x88:
                    
                    m_itemNum = value;
                    
                    retVal = true;
                    
                    break;
                
                default:
                    
                    break;
            }
            
            return retVal;
        }
        
        uint16_t GetType()
        {
            return m_itemNum;
        }
    };
    
    typedef list<Item> ItemList;
    
// ================================================
    
    typedef uint16_t Map8Val;
    
    //! \brief      Numerical reference to one of the Map16Tile objects
    //!             found in ZeldaGame::map16to8
    typedef uint16_t Map16Val;
    
    
    typedef uint16_t Map32Val;
    
// ================================================

    class Map16Tile
    {
    public:
        
        class Map16Exception : exception
        {
            
        };
        
        TileEntry upperLeft;
        TileEntry upperRight;
        TileEntry lowerLeft;
        TileEntry lowerRight;
        
        /// Map16Tile()
        
        Map16Tile()
        {
            
        }
        
        /// is this constructor useless?
        Map16Tile(TileEntry ul, TileEntry ur, TileEntry ll, TileEntry lr)
        
        : upperLeft(ul), upperRight(ur), lowerLeft(ll), lowerRight(lr) 
        {
            
        }
        
        TileEntry & operator[] (int i)
        {
            switch(i)
            {
                default:
                    throw Map16Exception(); break;
                case 0:
                    return upperLeft; break;
                case 1:
                    return upperRight; break;
                case 2:
                    return lowerLeft; break;
                case 3:
                    return lowerRight; break;
            }
        }
    };
    
// ================================================
    
    // I know the structure is identical to Map16Tile, but in the interest of
    // avoiding more Templates to Seth to process (at this time), I've declined
    // to make this code super generic just yet -MON
    class Map32Tile
    {
    public:
        
        class Map32TileException : exception
        {
            
        };
        
        Map16Val upperLeft;
        Map16Val upperRight;
        Map16Val lowerLeft;
        Map16Val lowerRight;
        
        Map32Tile(Map16Val ul, Map16Val ur, Map16Val ll, Map16Val lr)
        
        : upperLeft(ul), upperRight(ur), lowerLeft(ll), lowerRight(lr)
        {
            
        }
        
        Map16Val & operator[] (int i)
        {
            switch(i)
            {
                default:
                    throw Map32TileException(); break;
                case 0:
                    return upperLeft; break;
                case 1:
                    return upperRight; break;
                case 2:
                    return lowerLeft; break;
                case 3:
                    return lowerRight; break;
            }
        }
        
        Map16Val operator[] (int i) const
        {
            switch(i)
            {
                default:
                    throw Map32TileException(); break;
                case 0:
                    return upperLeft; break;
                case 1:
                    return upperRight; break;
                case 2:
                    return lowerLeft; break;
                case 3:
                    return lowerRight; break;
            }
        }
    };
    
// ================================================

    typedef Slot<Map16Val> Map16Slot;

    class Map16Buf
    {
    protected:
    
        bool m_large;
        
        Matrix<Map16Slot> m_data;
        
        // ================================================
    
    public:
        
        class Problem : public exception
        {
            
        };
        
        Map16Buf(bool large = false) : m_large(large), 
                                       m_data( large ? 64 : 32, large ? 64 : 32)
        {
            for(int i = 0; i < 20; ++i)
                i += 4;
        }
        
        Map16Buf(Map16Buf const &mb) : m_large(mb.m_large), m_data(mb.m_data)
        {
            for(int i = 0; i < 20; ++i)
                i += 4;
        }
        
        int Width()
        {
            return m_data.m_columns;
        }
        
        int Height()
        {
            return m_data.m_rows;
        }
        
        Map16Slot& operator () (uint8_t x, uint8_t y)
        {
            return m_data(y, x);
        }
        
        Map16Slot& operator () (uint16_t addr)
        {
            uint8_t x = GetX(addr) >> 4;
            uint8_t y = GetY(addr) >> 4;
            
            return m_data(y, x);
        }
        
        Map16Buf& operator = (Map16Buf const &rhs)
        {
            if( &rhs != this )
            {
                m_large = rhs.m_large;
                m_data  = rhs.m_data;
            }
            
            return (*this);
        }

    // ================================================

        static uint16_t GetX(uint16_t addr)
        {
            return (addr & 0x007E) << 3;
        }
        
        static uint16_t GetY(uint16_t addr)
        {
            return (addr & 0x1F80) >> 3;
        }
        
    // ================================================
    
    };
    
// ================================================
    
    struct Area
    {
        MetaData metaData;
        
        // indicates whether the area is large (1024x1024 pixels) or small (512x512 pixels)
        bool m_largeArea;
        
        // relevant for dealing with original game data. Not as relevant if you want to grossly
        // modify how the game works though.
        bool m_darkWorld;
        
        Ticket<BgTileMain> bgTileThemeMain;
        Ticket<BgTileAux>  bgTileThemeAux;
        
        Ticket<SprTileMain> sprTileThemeMain[3];
        Ticket<SprTileAux>  sprTileThemeAux;
        
        Ticket<OverBgTheme> bgPalTheme;
        Ticket<OverSprTheme> sprPalTheme[3];
        
        EntranceList entr;
        EntranceList holes;
        
        ExitList exits;
        
        // indicates whether falling rocks are present in this area for each of the 3 phases of the game.
        bool fallingRocks[0x03];
        
        SpriteList spr[0x03];
        
        ItemList items[0x03];
        
        Map16Buf areaMap;
        
        Area()
        {
            
        }
        
        ~Area()
        {
            
        }
    };
    
    typedef deque<Area> AreaDeque;
    
// ================================================

}

#endif