    
    #include <deque>
    #include <list>
    #include <utility>
    
    #include "containers.h"
    
#ifndef TEXTTYPES_H

    #define TEXTTYPES_H

namespace Text
{
    using std::deque;
    using std::exception;
    using std::list;
    
    using namespace Containers;
    
    class Problem : exception
    {
        
    };
    
    enum EntityType
    {
        Character = 0,
        NextPic   = 0x67,
        Choose,
        Item,
        Name,
        Window,
        Number,
        Position,
        ScrollSpd,
        SelChng,
        Crash,
        Choose2,
        Choose3,
        Scroll,
        Line1,
        Line2,
        Line3,
        Color,
        Wait,
        Sound,
        Speed,
        Command7b,
        Command7c,
        Command7d,
        WaitKey,
        End
    };
    
    class Entity
    {
        
    protected:
        
        EntityType m_et;
        
        Slot<uint8_t> m_arg;
        
    public:
        
        // Main Constructor_________________________________________
        Entity(EntityType p_et = Character,
               uint8_t p_arg = 0)
        
        : m_et(p_et)
        {
            uint8_t arg = 0;
            
            // -----------------------------
            
            switch(m_et)
            {
                // Command entities that don't take a parameter. 
                default:
                    
                    break;
                
                // Command entities that take a parameter.
                case Window:
                case Number:
                case Position:
                case ScrollSpd:
                case Color:
                case Wait:
                case Sound:
                case Speed:
                    
                    m_arg = p_arg;
                    
                    break;
                
                case Character:
                    
                    if(p_arg >= 0x67)
                    {
                        throw Problem();
                    }
                    
                    m_arg = p_arg;
                    
                    break;
            }
        }
        
        // Copy Constructor ________________________________________
        Entity(Entity const &rhs)
        {
            if(this != &rhs)
            {
                m_et = rhs.m_et;
                
                m_arg = (Slot<uint8_t>) rhs.m_arg;
            }
        }
        
        // Assignment Overload _____________________________________
        Entity& operator = (Entity &rhs)
        {
            if(this != &rhs)
            {
                m_et = rhs.m_et;
                
                m_arg = rhs.m_arg;
            }
            
            return (*this);
        }
        
        // _________________________________________________________
        EntityType GetType() const
        {
            return m_et;
        }
        
        // _________________________________________________________
        Slot<uint8_t> GetParam() const
        {
            return m_arg;
        }
    };
    
    typedef list<Entity> Message;
    
    typedef deque< Show<Message> > MessageDeque;
}

#endif