/*
 *	Compresch - compression library
 *  Copyright 2008 Disch
 *  See license.txt for details
 */


#ifndef __BLOCK_COPY_H__
#define __BLOCK_COPY_H__

/*
 *	Copy -- Not even really compression, just a block copy from the source buffer
 *          to the decompression buffer.
 *
 */


class Block_Copy : public Block
{
        
    // ==============================================
        
    protected:
        
	    uint8_t dat;
        
    public:
        
    // ==============================================
        
        virtual int GetBodySize()
        {
            return 1;
        }
        
    // ==============================================
        
        virtual int GetMinLength()
        {
            return 1;
        }
        
    // ==============================================
        
        virtual int Output(uint8_t* buf) { buf[0] = dat; return 1; }
        
    // ==============================================
        
        virtual int Shrink(int newstart,int newstop)
	    {
		    start = newstart;
		    stop = newstop;
		    len = stop - start;
		    return 0;
	    }
        
    // ==============================================
        
        virtual Block* Dup()	{	return (new Block_Copy(*this));		}
        
    // ==============================================
        
    /* Build all possible blocks */
        
        static void	Build(uint8_t type, BlockList& lst, const uint8_t* src, int srclen)
	    {
		    uint8_t run;

		    int i   = 0, j = 0;
		    int len = 0;

            Block_Copy*		bk;

            // ------------------------

            // Do nothing.
	    }
        
    // ==============================================

	    /* Decompression */
	    virtual int Decompress(int len, const uint8_t *src, uint8_t* dst, int &dstpos)
	    {
            int i = 0;

            // ---------------------------

		    if(dst)
            {
                for(i = 0; i < len; ++i)
				    dst[dstpos++] = src[i];
            }
		    else
			    dstpos += len;

            // since it's a direct copy of a string of bytes,
            // we have to advance the source pointer by this much.
		    return len; 
	    }
};

// ***********************************************************************
      


#endif