#-------------------------------------------------------------------------------
# Name:        thorn_helpers
# Purpose:     
#
# Author:      User
#
# Created:     30/03/2012
# Copyright:   (c) User 2012
# Licence:     <your licence>
#-------------------------------------------------------------------------------
#!/usr/bin/env python

from collections import OrderedDict

# ==============================================================================

class obj_def:
    
    # __________________________________________________________________________
    def __init__(self, name, params = None, block = None):
        
        self.name   = name
        self.block  = block
        self.params = OrderedDict( [ (k.name, k.typ) for k in params ] )
        
        # Can't have repeated param names, even if they're different types
        if len(self.params) < len(params):
            
            raise Exception()
        
    # __________________________________________________________________________
    def __str__(self):
        
        params_str_list = [ str(p) for p in self.params]
        
        params_str = ", ".join(params_str_list)
        
        return '%s(%s)' % (self.name, params_str)
    
    # __________________________________________________________________________
    def get_signature(self):
        
        return self.params.values()

    # __________________________________________________________________________
    def get_param_names(self):
        
        return self.params.keys()
    
# ==============================================================================

class obj_param:
    
    def __init__(self, typ, name):
        
        self.typ  = typ
        self.name = name
        
    # __________________________________________________________________________
    def __str__(self):
        
        return '%s %s' % (self.typ, self.name)

# ==============================================================================

class block_def:
    
    def __init__(self, name, params, indices):
        
        self.name    = name
        self.params  = params
        self.indices = indices
    
    # __________________________________________________________________________
    def __str__(self):
        
        indices_str = ", ".join([str(i) for i in self.indices])
        
        return '%s(0x%04x, %i, %i) { %s };' % (self.name,
                                               self.params[0].obj,
                                               self.params[1].obj,
                                               self.params[2].obj,
                                               indices_str) 
        
# ==============================================================================

class indir_block_def:
    
    def __init__(self, name, offset, width, height, indices):
        
        self.name   = name
        self.offset = offset
        self.width  = width
        self.height = height
        
        self.indices = indices
        
    # __________________________________________________________________________
    def __str__(self):
        
        indices_str = ", ".join([str(i) for i in self.indices])
        
        return '%s(0x%04x, %i, %i) { %s };' % (self.name,
                                               self.params[0].obj,
                                               self.params[1].obj,
                                               self.params[2].obj,
                                               indices_str) 
    
    # __________________________________________________________________________
    def cpp_str(self):
        
        indices_str = ", ".join([str(i) for i in self.indices])
        
        fmt_str = '{ "%s", IndirectBlock(0x%04x, %i, %i, { %s }) }'
        
        return fmt_str % (self.name,
                          self.offset,
                          self.height,
                          self.width,
                          indices_str)

# ==============================================================================

class obj_call:
    
    def __init__(self, name, params = None):
        
        self.name   = name
        self.params = params
        
    # __________________________________________________________________________
    def __str__(self):
        
        s = str(self.name)
        
        for p in self.params:
            
            s += str(p)
            
        return s
        
# ==============================================================================

class assign:
    
    def __init__(self, lhs, rhs):
        
        self.lhs = lhs
        self.rhs = rhs
        
# ==============================================================================

class var_decl:
    
    def __init__(self, typ, name, rhs):
        
        self.typ  = typ
        self.name = name
        self.rhs  = rhs
        
# ==============================================================================

class expr:
    
    def __init__(self, left, op = None, right = None):
        
        self.left  = left
        self.op    = op
        self.right = right
        
        if (left and right) and not op:
            raise Exception
    
    # __________________________________________________________________________
    def __str__(self):
        
        # could probably be done more elegantly, but for right now this works.
        s = "(%s" % self.left
        
        if self.op:
            
            s += " %s " % (self.op)
            
            if self.right:
                
                s += "%s" % (self.right)
        
        s += ")"
        
        return s
        
# ==============================================================================

class node:
    
    def __init__(self, typ, obj):
        
        self.typ = typ
        self.obj = obj
    
    def __str__(self):
        
        return "node type: %s, node obj %s" % (self.typ, self.obj)
    
# ==============================================================================

class conditional_block:
    
    def __init__(self, word, cond_expr, block):
        
        self.word      = word
        self.cond_expr = cond_expr
        self.block     = block
        
# ==============================================================================

class loop_block:
    
    def __init__(self, loop_expr, block):
        
        self.loop_expr = loop_expr
        self.block     = block
        
# ==============================================================================

class block_ref:
    
    def __init(self, block, hflip = False, vFlip = False):
        
        self.block = block
        self.hflip = hflip
        self.vflip = vflip
        
    # __________________________________________________________________________
    def hflip(self):
    
        self.hflip = not self.hflip
        
    # __________________________________________________________________________
    def vflip(self):
        
        self.vflip = not self.vflip
        
       
if __name__ == '__main__':
    pass
