#-------------------------------------------------------------------------------
# Name:        thorn_lex
# Purpose:     
#
# Author:      User
#
# Created:     30/03/2012
# Copyright:   (c) User 2012
# Licence:     <your licence>
#-------------------------------------------------------------------------------
#!/usr/bin/env python

import ply.lex as lex

# ==============================================================================

class thornLexer:

    reserved = [
                'bg1',
                'bg2',
                'bgx',
                'true',
                'false',

                'pos',
                'x',
                'y',

                'chr',
                'eq',
                'not_eq',

                'block',
                'width',
                'height',
                'hflip',
                'vflip',
                'vhflip',
                
                'obj',
                
                'if',
                'elif',
                'else',
                'unless',
                
                'in',
                'tuple',
                
                'loop'
               ]

    single_ops ={
                    '+' : 'PLUS',
                    '-' : 'MINUS',
                    '*' : 'MULTIPLY',
                    '/' : 'DIVIDE',
                    '(' : 'LPAREN',
                    ')' : 'RPAREN',
                    '{' : 'LBRACKET',
                    '}' : 'RBRACKET',
                    '.' : 'PERIOD',
                    ',' : 'COMMA',
                    ';' : 'SEMI',
                    '=' : 'ASSIGN'
                }
    
    double_ops ={
                    '+=' : 'PLUS_ASSIGN',
                    '-=' : 'MINUS_ASSIGN',
                    '*=' : 'MULTIPLY_ASSIGN',
                    '/=' : 'DIVIDE_ASSIGN',
                    '==' : 'LOG_EQ',
                    '!=' : 'LOG_NEQ'
                }
    
    types = [
                'repeat',
                'chr_t',
                'block_t',
                'offset'
            ]
    
    tokens = [
                'NUMBER',
                'HEXNUMBER',

                'TYPE',

                'ID'
             ]
    
    tokens += single_ops.values()
    tokens += double_ops.values()
    tokens += [ i.upper() for i in reserved]
    
    # __________________________________________________________________________
    def advance_column(self, t):
        
        t.column_number = t.lexer.column_number
        
        t.lexer.column_number += len(t.value)
        
    # __________________________________________________________________________
    def t_newline(self, t):
        r'\n+'
        
        t.lexer.lineno += len(t.value);
        
        # reset the column counter
        t.lexer.column_number = 0
        
    # __________________________________________________________________________
    def t_carriage_return(self, t):
        '[\r]+'
        
        self.advance_column(t)
        
        pass
    
    # __________________________________________________________________________
    def t_space(self, t):
        '[ ]+'
        
        self.advance_column(t)
        
        pass

    # __________________________________________________________________________
    def t_tabs(self, t):
        '[\t]+'
        
        self.advance_column(t)
        
        pass
    
    # __________________________________________________________________________
    def t_comment(self, t):
        r'\#.*'
        
        self.advance_column(t)
        
        pass

    # __________________________________________________________________________
    def t_error(self, t):

        print "Invalid token: %s" % (t)

        self.advance_column(t)
        
        t.lexer.skip(1)
        
        pass
    
    # __________________________________________________________________________
    def t_ID(self, t):
        r'[_a-zA-Z]\w*'
        
        self.advance_column(t)
        
        token = t.value
        
        if token in self.types:
            t.type  = 'TYPE'
        elif token in self.reserved:
            t.type  = token.upper()
        else:
            t.type  = 'ID'
        
        return t

    # __________________________________________________________________________
    def t_HEXNUMBER(self, t):
        r'0x[a-fA-F0-9]+'
        
        self.advance_column(t)
        
        t.value = int(t.value[2:], 16)
        t.type  = 'NUMBER'
        
        return t
        
    # __________________________________________________________________________
    def t_NUMBER(self, t):
        r'[0-9]+'
        
        self.advance_column(t)
        
        t.value = int(t.value)
        
        return t
    
    def t_double_op(self, t):
        r'[\+\-\*\/\!\=]\='
        
        self.advance_column(t)
        
        t.type = self.double_ops[t.value]
        
        return t
    
    # __________________________________________________________________________
    def t_single_op(self, t):
        r'[\+\-\*\/\.\,\;\=\(\)\{\}]'
        
        self.advance_column(t)
        
        t.type = self.single_ops[t.value]
        
        return t
    
    # __________________________________________________________________________
    def __init__(self, **kwargs):
        
        self.lexer = lex.lex(module = self, **kwargs)
        
        self.lexer.column_number = 0

# ==============================================================================

if __name__ == '__main__':
    pass
