#-------------------------------------------------------------------------------
# Name:        thorn_yacc
# Purpose:     
#
# Author:      User
#
# Created:     30/03/2012
# Copyright:   (c) User 2012
# Licence:     <your licence>
#-------------------------------------------------------------------------------
#!/usr/bin/env python

import ply.yacc as yacc

from thorn_lex import thornLexer

# usually bad practice, but we need them all
from thorn_helpers import *

class thornParser:
    
    tokens = thornLexer.tokens
    
    precedence =    (
                        ('left',  'PLUS',   'MINUS'),
                        ('left',  'DIVIDE'),
                        ('left', 'PERIOD'),
                        ('left',  'LPAREN', 'RPAREN'),
                        ('right', 'UNARY_MINUS'),
                        ('nonassoc',    'EXPRESSION'),
                    )
    
    # Just explicitly stating this, otherwise the first encountered
    # parse symbol will be the top level node type. We already have module
    # as the first one, but I'd like to make it explicit in case code gets 
    # shifted around at some point on accident.
    start = 'module'
    
    # __________________________________________________________________________
    def __init__(self, **kwargs):
        
        self.parser = yacc.yacc(module = self, **kwargs)
        
    # __________________________________________________________________________
    def p_module(self, p):
        '''
        module : object_def
               | block_def
        '''
        # modules are simple lists
        p[0] = [ p[1] ]
        
    # __________________________________________________________________________
    def p_module_add(self, p):
        '''
        module  : module object_def
                | module block_def
        '''
        
        p[1].append(p[2])
        
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_object_def_noparams(self, p):
        '''
        object_def : OBJ ID LPAREN RPAREN obj_block
        '''
        
        p[0] = node('obj_def', obj_def(p[2], [], p[5]) )
        
    # __________________________________________________________________________
    def p_object_def_withparams(self, p):
        '''
        object_def : OBJ ID LPAREN obj_params RPAREN obj_block
        '''
        p[0] = node('obj_def', obj_def(p[2], p[4], p[6]) )
        
    # __________________________________________________________________________
    def p_obj_params_single(self, p):
        '''
        obj_params : obj_param
        '''
        
        # an obj_params node is a simple list
        p[0] = [ p[1] ]
        
    # __________________________________________________________________________
    def p_obj_params_add(self, p):
        '''
        obj_params : obj_params COMMA obj_param
        '''
        p[1].append(p[3])
        
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_obj_param(self, p):
        '''
        obj_param : TYPE ID
                  | BLOCK ID
        '''
        
        if p[1] == 'block':
            p[0] = obj_param('block_ref', p[2])
        else:
            p[0] = obj_param(p[1], p[2])
        
    # __________________________________________________________________________
    def p_obj_block_empty(self, p):
        '''
        obj_block : LBRACKET RBRACKET
        '''
        p[0] = []
        
    # __________________________________________________________________________
    def p_obj_block_nonempty(self, p):
        '''
        obj_block : LBRACKET obj_statements RBRACKET
        '''
        # obj_block nodes are simple list of obj_statement nodes
        # of various types indicated by the tag in the first entry
        # of their tuple.
        p[0] = p[2]
        
    # __________________________________________________________________________
    def p_obj_statements_single(self, p):
        '''
        obj_statements : obj_statement SEMI
                       | loop_block
                       | conditional_block
        '''
        # obj_statements is a glue node in the parser that ultimately
        # is identical to the list that the obj_block represents
        # It only functions as an intermediary, and doesn't have its own
        # node type.
        p[0] = [ p[1] ]
        
    # __________________________________________________________________________
    def p_obj_statements_add(self, p):
        '''
        obj_statements : obj_statements obj_statement SEMI
                       | obj_statements loop_block
                       | obj_statements conditional_block
        '''
        p[1].append(p[2])
        
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_obj_statement(self, p):
        '''
        obj_statement : obj_call
                      | assign
                      | var_decl
        '''
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_obj_call(self, p):
        '''
        obj_call : ID expr_tuple
        '''
        p[0] = node('obj_call', obj_call(p[1], p[2]) )
        
    # __________________________________________________________________________
    def p_expr_tuple_empty(self, p):
        '''
        expr_tuple : LPAREN RPAREN
        '''
        p[0] = []
        
    # __________________________________________________________________________
    def p_expr_tuple(self, p):
        '''
        expr_tuple : LPAREN expr_tuple_elements RPAREN
        '''
        # expr_tuple nodes are simple lists of expressions
        p[0] = p[2]
        
    # __________________________________________________________________________
    def p_expr_tuple_elements_single(self, p):
        '''
        expr_tuple_elements : expr
        '''
        p[0] = [ p[1] ]
        
    # __________________________________________________________________________
    def p_expr_tuple_elements_add(self, p):
        '''
        expr_tuple_elements : expr_tuple_elements COMMA expr
        '''
        p[1].append(p[3])
        
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_expr_standalone(self, p):
        '''
        expr : ID
        '''
        p[0] = node('identifier', p[1])
    
    # __________________________________________________________________________
    def p_node_int(self, p):
        '''
        expr : NUMBER
        '''
        p[0] = node('number', p[1])
        
    # __________________________________________________________________________
    def p_node_xy(self, p):
        '''
        expr : X
             | Y
        '''
        p[0] = node('offset', p[1])
        
    # __________________________________________________________________________
    def p_node_block_dimension(self, p):
        '''
        expr : WIDTH
             | HEIGHT
        '''
        p[0] = node('block_dimension', p[1])
    
    # __________________________________________________________________________
    def p_node_block_flip(self, p):
        '''
        expr : HFLIP
             | VFLIP
             | VHFLIP
        '''
        p[0] = node('block_flip', p[1])
    
    # __________________________________________________________________________
    def p_node_chr(self, p):
        '''
        expr : CHR
        '''
        p[0] = node('chr', p[1])
        
    # __________________________________________________________________________
    def p_node_truth_value(self, p):
        '''
        expr : TRUE
             | FALSE
        '''
        p[0] = node('truth_value', p[1])
        
    # __________________________________________________________________________
    def p_expr_background(self, p):
        '''
        expr : BG1
             | BG2
             | BGX
        '''
        p[0] = node('background', p[1])
    
    # __________________________________________________________________________
    def p_expr(self, p):
        '''
        expr : expr PLUS expr
             | expr MINUS expr
             | expr MULTIPLY expr
             | expr DIVIDE expr
             | expr PERIOD expr
             | expr LOG_EQ expr
             | expr NOT_EQ expr
        '''
        p[0] = node('expr', expr(p[1], p[2], p[3]) )
        
        
    # __________________________________________________________________________
    def p_assign(self, p):
        '''
        assign : expr ASSIGN expr
        '''
        
        if p[1].typ not in [ 'background',
                             'identifier',
                             'chr',
                             'offset' ]:
            raise Exception
        
        p[0] = node('assign', assign(p[1], p[3]) )
        
    # __________________________________________________________________________
    def p_expr_compound_assign(self, p):
        '''
        assign : expr PLUS_ASSIGN expr
               | expr MINUS_ASSIGN expr
               | expr MULTIPLY_ASSIGN expr
               | expr DIVIDE_ASSIGN expr
        '''
        
        if p[1].typ not in [ 'background',
                             'identifier',
                             'chr',
                             'offset' ]:
            raise Exception
        
        # The operation is the truncation of the '=' portion
        # e.g. '+=' becomes '+'
        op = p[2][0:1]
        
        rhs = node('expr', expr(p[1], op, p[3]) )
        
        p[0] = node('assign', assign(p[1], rhs) )
        
    # __________________________________________________________________________
    def p_var_decl(self, p):
        '''
        var_decl : TYPE ID ASSIGN expr
                 | BLOCK ID ASSIGN expr
        '''
        if p[1] == 'block':
            p[1] = 'block_ref'
        
        p[0] = node('var_decl', var_decl(p[1], p[2], p[4]) )
        
    # __________________________________________________________________________
    def p_expr_negate(self, p):
        '''
        expr : MINUS expr %prec UNARY_MINUS
        '''
        p[0] = node('unary_minus', p[2])
        
    # __________________________________________________________________________
    def p_expr_parentheses(self, p):
        '''
        expr : LPAREN expr RPAREN  %prec EXPRESSION
        '''
        p[0] = p[2]
        
    # __________________________________________________________________________
    def p_expr_in(self, p):
        '''
        expr : expr IN TUPLE expr_tuple
        '''
        
        p[0] = node('expr',
                    expr(p[1],
                         p[2],
                         node('tuple', p[4]) ) )
        
    # __________________________________________________________________________
    def p_block(self, p):
        '''
        block_def : BLOCK ID expr_tuple indices_block SEMI
        '''
        if len(p[3]) in [ 2, 3 ]:
            p[0] = node('block_def', block_def(p[2], p[3], p[4]) )
        else:
            raise Exception
        
    # __________________________________________________________________________
    def p_indices_block_none(self, p):
        '''
        indices_block : empty
        '''
        p[0] = []
        
    # __________________________________________________________________________
    def p_indices_block_empty(self, p):
        '''
        indices_block : LBRACKET RBRACKET
        '''
        p[0] = []
        
    # __________________________________________________________________________
    def p_indices_block(self, p):
        '''
        indices_block : LBRACKET indices_block_entries RBRACKET
        '''
        p[0] = p[2]
        
    # __________________________________________________________________________
    def p_indices_block_entries_single(self, p):
        '''
        indices_block_entries : NUMBER
        '''
        p[0] = [ p[1] ]
        
    # __________________________________________________________________________
    def p_indices_block_entries_add(self, p):
        '''
        indices_block_entries : indices_block_entries COMMA NUMBER
        '''
        p[1].append(p[3])
        
        p[0] = p[1]
        
    # __________________________________________________________________________
    def p_conditional_block(self, p):
        '''
        conditional_block : UNLESS expr_tuple obj_block
                          | IF     expr_tuple obj_block
        '''
        
        if len(p[2]) != 1:
            raise Exception('Loop block requires exactly 1 expression to'
                            'evaluate')
        
        cond = p[2][0]
        
        p[0] = node('conditional_block', conditional_block(p[1], cond, p[3]) )
        
    # __________________________________________________________________________
    def p_loop(self, p):
        '''
        loop_block : LOOP expr_tuple obj_block
        '''
        
        if len(p[2]) != 1:
            raise Exception('Loop block requires exactly 1 expression to'
                            'evaluate')
        
        arg = p[2][0]
        
        p[0] = node('loop_block', loop_block(arg, p[3]) )
        
    # __________________________________________________________________________
    def p_error(self, p):
        
        # Just discard the token and tell the parser it's okay.
        print "Syntax error at token", p.type
        
        print p
        
        # print self.parser.symstack
        
        self.parser.errok()
        
    # __________________________________________________________________________
    def p_empty(self, p):
        '''
        empty :
        '''
        pass
        
    def p_empty2(self, p):
        '''
        murp : empty
        '''
        pass

if __name__ == '__main__':
    pass
