# -*- coding: utf-8 -*-
 
import sys
import random
import json
import blackmagic as bm
from PySide.QtCore import *
from PySide.QtGui import *
from cStringIO import StringIO

from zgame import *
from textmode import convert_char
from mapview import MapView
from zeldaview import Ui_Zelda3MainWindow

#For fun and profit
NonEmptyLayers = [0, 2, 3, 5, 7, 10, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 26, 27, 29, 30, 34, 37, 40, 41, 42, 43, 44, 45, 46, 47, 48, 50, 51, 52, 53, 55, 58, 59, 60, 63, 64, 66, 67, 69, 71, 74, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 90, 91, 93, 94, 98, 101, 104, 105, 106, 107, 108, 109, 110, 111, 112, 114, 115, 116, 117, 119, 122, 123, 124, 127, 128, 136, 147, 148, 149, 150, 151, 156, 157, 158, 159]



#Specify our "MainWindow" class from the generated code
class MyMainWindow(QMainWindow):
  def __init__(self, parent=None):
    #Load the ui
    super(MyMainWindow, self).__init__(parent)
    self.ui = Ui_Zelda3MainWindow()
    self.ui.setupUi(self)

    #The current game and map objects.
    self.zGame = None
    self.settingMap = False

    #Attempt to open the ROM
    self.ui.btn_dialogue_txt.clicked.connect(self.show_dlg_text)

    #Custom drawing code
    self.ui.btn_test_drawing.clicked.connect(self.custom_draw)
    self.ui.btn_prev_map.clicked.connect(self.prevMap)
    self.ui.btn_next_map.clicked.connect(self.nextMap)

    #Map combo box
    self.ui.cmb_curr_map.currentIndexChanged.connect(self.newCurrMapFromCombo)

    #Hook up signals: layer visibility
    self.ui.tbtn_map_layer.clicked.connect(self.showHideMapLayer)
    self.ui.tbtn_annot_layer.clicked.connect(self.showHideAnnotLayer)

    #Signal for setting the map size
    self.ui.tbtn_misc_layer.clicked.connect(self.showTransitionAreas)

    #Menu hookups
    self.ui.action_Open.triggered.connect(self.load_rom)


  def ConvertBgToARGB(self, bgcolors):
    for bg in bgcolors:
      color = bgcolors[bg]
      if color[:2]=='0x':
        #Hex can be 32-bit or 13(?) bit
        colorRem = color[2:]
        if len(colorRem)==6:
          color = int(colorRem, 16)
        elif len(colorRem)==4:
          #Convert
          color = int(colorRem, 16)
          r = (color&0xF)
          g = ((color>>5)&0x1F)
          b = ((color>>10)&0xF)
          color = (r<<(16+3)) | (g<<(8+3)) | (b<<(0+3))
        else:
          raise 'Unknown hex color depth'
      else:
        color = int(color)
      bgcolors[bg] = (0xFF000000 | color)

  def ParseConnectivity(self, annot, connectArray):
    if not connectArray: return

    #Convert from string notation
    annot.areaConnections = []
    for row in connectArray:
      annot.areaConnections.append(map((lambda x: int(x, 0)), row))

    #Build reverse lookup
    for row in xrange(len(connectArray)):
      for col in xrange(len(connectArray[row])):
        annot.areaLookup[int(connectArray[row][col], 0)] = (col, row)


  def LoadAnnotations(self, filename):
    res = Annotations()
    fileHandle = None
    jsonData = None
    try:
      fileHandle = open(filename, 'r')
      jsonData = json.load(fileHandle)
    except IOError: 
      return res

    #Respond to our json data
    try:
      #Convert all bgcolors to ARGB
      self.ConvertBgToARGB(jsonData["bgcolors"])

      #Parse all overworld area metadata
      for overID in jsonData["overworld"]:
        overData = jsonData["overworld"][overID]
        if overData.get("name"):
          res.areaNames[int(overID)] = overData["name"]
        if overData.get("bgcolor"):
          res.areaBgColors[int(overID)] = jsonData["bgcolors"][overData["bgcolor"]]

      #Parse connectivity
      self.ParseConnectivity(res, jsonData["lightworld_connectivity"])
      #self.ParseConnectivity(res, jsonData["darkworld_connectivity"])  #TODO: Currently, this will overwrite...

    except KeyError, ValueError:
      res = Annotations() #Clear it
      print 'Error loading annotations; skipping...'

    #Done
    fileHandle.close()
    return res

       
  def load_rom(self):
    #Prompt for a ROM file (for now Unicode won't work)
    fileName = QFileDialog.getOpenFileName(self, "Open Zelda3 ROM", "zelda3.sfc", "Super Famicon (*.sfc)")[0]
    if not fileName:
      return

    #Load our default set of annotations
    annotations = self.LoadAnnotations("data/overworld.json.txt")

    #Load it via our ExpandInfo class
    fileExInfo = bm.MakeExpandInfo(str(fileName))
    fileLoader = bm.RomLoader(fileExInfo)

    #Now extract some things from it
    self.zGame = ZGame(bm.ZeldaGame(), annotations)
    self.ui.mv_overworld.setGame(self.zGame)
    fileLoader.LoadGame(self.zGame.game)
    self.ui.list_dlg_msgs.clear()
    self.ui.cmb_curr_map.clear()

    #Some resetting and cleanup
    self.show_dlg_text()

  def show_dlg_text(self):
    if self.ui.list_dlg_msgs.count()==0 and self.zGame:
      #Rebuild the message list (or try to) if the list is empty
      for msg in self.zGame.game.messages:
        buff = StringIO()
        buff.write("[" + hex(self.ui.list_dlg_msgs.count()) + "]  ")
        for packet in msg:
          if packet.type==bm.TextEntityType.Character:
            buff.write(convert_char(packet.param.to_i()))
          else:
            buff.write("[" + str(packet.type))
            if not packet.param.IsEmpty():
              buff.write(" " + str(packet.param.to_i()))
            buff.write("]")
        self.ui.list_dlg_msgs.addItem(buff.getvalue())
    self.ui.sw_curr_tab.setCurrentIndex(0)

  def custom_draw(self):
    #Add items if needed
    if self.zGame and self.ui.cmb_curr_map.count()==0:
      self.settingMap = True #Grumble
      for i in xrange(self.zGame.getNumberOfMaps()):
        self.ui.cmb_curr_map.addItem(self.zGame.getAreaName(i))
      self.settingMap = False

    #Pick a random number to show
    mapID = NonEmptyLayers[random.randint(0, len(NonEmptyLayers)-1)]
    self.ui.cmb_curr_map.setCurrentIndex(-1)

    #Refresh (should be done automatically)
    self.showHideMapLayer()
    self.showHideAnnotLayer()
    self.showTransitionAreas()
    self.ui.sw_curr_tab.setCurrentIndex(1)

    #This has to be done after?
    self.ui.cmb_curr_map.setCurrentIndex(mapID)

  def newCurrMapFromCombo(self):
    if not self.zGame:
      return
    self.newMap(self.ui.cmb_curr_map.currentIndex())

  def showHideMapLayer(self):
    isChecked = self.ui.tbtn_map_layer.isChecked()
    self.ui.mv_overworld.showHideMapLayer(isChecked)

  def showHideAnnotLayer(self):
    isChecked = self.ui.tbtn_annot_layer.isChecked()
    self.ui.mv_overworld.showHideAnnotLayer(isChecked)

  def showTransitionAreas(self):
    isChecked = self.ui.tbtn_misc_layer.isChecked()
    self.ui.mv_overworld.showBorderAreas(isChecked)

  def prevMap(self):
    if not self.zGame:
      self.ui.mv_overworld.triCountAdd(-1)
      return
    self.newMap(self.ui.cmb_curr_map.currentIndex() - 1)

  def nextMap(self):
    if not self.zGame:
      self.ui.mv_overworld.triCountAdd(1)
      return
    self.newMap(self.ui.cmb_curr_map.currentIndex() + 1)

  def newMap(self, mapID):
    if self.settingMap:
      return
    self.settingMap = True
    if self.zGame.setCurrMap(mapID):
      self.ui.cmb_curr_map.setCurrentIndex(mapID)
      self.ui.mv_overworld.newMap(mapID)
    self.settingMap = False

  def MakeButtonWithIcon(self, path, checkable):
    icn = QIcon(path)
    res = QPushButton(icn, '')
    res.setIconSize(icn.availableSizes()[0])
    res.setCheckable(checkable)
    return res

