#include <cmath>
#include <boost/python.hpp>

#include "graphics_helpers.hpp"

using namespace boost::python;
using namespace Graphics;



//Export all relevant items in the Graphics namespace
void export_graphics()
{
  class_<ColorTile>("ColorTile", init<Tile<Pixel3bpp> &, Palette<8> &, int>())
    .def("get", &ColorTile::operator(), return_value_policy<reference_existing_object>())
  ;

  class_<TileSet<Pixel3bpp, 0x40> >("TileSet3bpp") //Pixel3bpp == 3
    .def("NumTiles", &TileSet<Pixel3bpp, 0x40>::NumTiles)
    .def("get", &TileSet<Pixel3bpp, 0x40>::operator(), return_value_policy<reference_existing_object>())
  ;

  class_<Containers::Ticket<TileSet3bpp>,
         Containers::Ticket<TileSet3bpp>*,
         boost::noncopyable>("BgTileSetTicket")
    .def("Redeem", &Containers::Ticket<TileSet3bpp>::Redeem)
  ;


  class_<TileTheme<TileSet3bpp, 8> >("BgTileMainTheme")
    .def("get", &TileTheme<TileSet3bpp, 8>::operator [], return_value_policy<reference_existing_object>())
  ;
  class_<TileTheme<TileSet3bpp, 4> >("BgTileAuxTheme")
    .def("get", &TileTheme<TileSet3bpp, 4>::operator [], return_value_policy<reference_existing_object>())
  ;

  //Used internally by Python
  class_<TileSet<Pixel3bpp, 0x200> >("BgTileSet")
    //Indexing support
    .def("__getitem__", &TileSet_Pixel3bpp_0x200__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &TileSet_Pixel3bpp_0x200__set_item__)
  ;

  class_<Tile<Pixel3bpp> >("Tile3Bpp")
    .def("get", &Tile<Pixel3bpp>::operator(), return_value_policy<reference_existing_object>())
  ;

  class_<Color>("Color")
    .def("rgb", getColor_rgb)
  ;

  //Various palettes
  class_<Palette<7> >("Palette7")
    .def("length", &Palette<7>::Length)
    .def("size", &Palette<7>::Size)

    //Indexing support
    .def("__getitem__", &Palette_7__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Palette_7__set_item__)
  ;
  class_<Palette<8> >("Palette8")
    .def("length", &Palette<8>::Length)
    .def("size", &Palette<8>::Size)

    //Indexing support
    .def("__getitem__", &Palette_8__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Palette_8__set_item__)
  ;


  //Various groups
  class_<Graphics::Group<8,16>,
         Graphics::Group<8,16>* >("Group_8_16")

    //Indexing support
    .def("__getitem__", &Group_8_16__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Group_8_16__set_item__)
  ;
  class_<Graphics::Group<7,5>,
         Graphics::Group<7,5>* >("Group_7_5")
    .def("PaletteSize", &Graphics::Group<7, 5>::PaletteSize)

    //Indexing support
    .def("__getitem__", &Group_7_5__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Group_7_5__set_item__)
  ;
  class_<Graphics::Group<7,3>,
         Graphics::Group<7,3>* >("Group_7_3")
    .def("PaletteSize", &Graphics::Group<7, 3>::PaletteSize)

    //Indexing support
    .def("__getitem__", &Group_7_3__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Group_7_3__set_item__)
  ;
  class_<Graphics::Group<7,1>,
         Graphics::Group<7,1>* >("Group_7_1")
    .def("PaletteSize", &Graphics::Group<7, 1>::PaletteSize)

    //Indexing support
    .def("__getitem__", &Group_7_1__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &Group_7_1__set_item__)
  ;


  class_<Containers::Ticket<Graphics::Group<7,5> >,
         Containers::Ticket<Graphics::Group<7,5> >*,
         boost::noncopyable>("Group_7_5Ticket")
    .def("Redeem", &Containers::Ticket<Graphics::Group<7,5> >::Redeem)
  ;
  class_<Containers::Ticket<Graphics::Group<7,3> >,
         Containers::Ticket<Graphics::Group<7,3> >*,
         boost::noncopyable>("Group_7_3Ticket")
    .def("Redeem", &Containers::Ticket<Graphics::Group<7,3> >::Redeem)
  ;
  class_<Containers::Ticket<Graphics::Group<7,1> >,
         Containers::Ticket<Graphics::Group<7,1> >*,
         boost::noncopyable>("Group_7_1Ticket")
    .def("Redeem", &Containers::Ticket<Graphics::Group<7,1> >::Redeem)
  ;


  class_<Graphics::OverBgTheme >("OverBgTheme")
    .def_readonly("primary", &Graphics::OverBgTheme::primary)
    .def_readonly("secondary1", &Graphics::OverBgTheme::secondary1)
    .def_readonly("secondary2", &Graphics::OverBgTheme::secondary2)
    .def_readonly("tertiary", &Graphics::OverBgTheme::tertiary)
  ;


  class_<Containers::Ticket<Graphics::BgTileMain>,   //Primary class specification
         Containers::Ticket<Graphics::BgTileMain>*,  //Also specify it can be a pointer (avoids "No to_python (by-value) converter")
         boost::noncopyable >("BgTileMainTicket")    //Also specify it has no copy constructor (avoids cryptic "Can't convert" error)
    .def("Redeem", &Containers::Ticket<Graphics::BgTileMain>::Redeem)
  ;


  class_<Containers::Ticket<Graphics::BgTileAux>,
         Containers::Ticket<Graphics::BgTileAux>*,
         boost::noncopyable >("BgTileAuxTicket")
    .def("Redeem", &Containers::Ticket<Graphics::BgTileAux>::Redeem)
  ;

  class_<Containers::Ticket<Graphics::OverBgTheme>,
         Containers::Ticket<Graphics::OverBgTheme>*,
         boost::noncopyable >("OverBgThemeTicket")
    .def("Redeem", &Containers::Ticket<Graphics::OverBgTheme>::Redeem)
  ;
}

