#include <cmath>
#include <boost/python.hpp>

#include "overworld_helpers.hpp"

using namespace boost::python;
using namespace Overworld;



//Export all relevant items in the Overworld namespace
void export_overworld()
{
  //PalRef enum
  enum_<PalRef>("PalRef")
    .value("pal0", pal0)
    .value("pal1", pal1)
    .value("pal2", pal2)
    .value("pal3", pal3)
    .value("pal4", pal4)
    .value("pal5", pal5)
    .value("pal6", pal6)
    .value("pal7", pal7)
  ;

  class_<Map16Tile>("Map16Tile")
    .def("get", &Map16Tile::operator [], return_value_policy<reference_existing_object>())
  ;

  class_<TileEntry>("TileEntry")
    .add_property("chr", &TileEntry::GetChr)
    .def_readonly("pal", &TileEntry::m_pal)  //This enum is used as an integer
    .def_readonly("v_flip", &TileEntry::m_vFlip)
    .def_readonly("h_flip", &TileEntry::m_hFlip)
    .def_readonly("priority", &TileEntry::m_priority)
  ;

  class_<Map16Buf>("Map16Buf", init<const Map16Buf&>())
    .def("width", &Map16Buf::Width)
    .def("height", &Map16Buf::Height)
    .def("get_slot", getMapSlot_1, return_value_policy<reference_existing_object>())  //I think existing makes sense here.
  ;

  class_< Position<16> >("Position16")
    .def("GetX", &Position<16>::GetX)
    .def("GetY", &Position<16>::GetY)
  ;

  class_<Sprite>("Sprite")
    .def_readonly("spriteNum", &Sprite::spriteNum)
    .def_readonly("pos", &Sprite::pos)
  ;

  class_< std::list<Sprite> >("SpriteList")
    .def("size", &std::list<Sprite>::size)

    //Enable iteration, but not random access
   .def("__iter__", iterator< std::list<Sprite> >())
  ;

  class_<Area>("Area")
    .def_readonly("large_area", &Area::m_largeArea)
    .def_readonly("dark_world", &Area::m_darkWorld)
    .def_readonly("area_map", &Area::areaMap)
    .def_readonly("bg_tile_theme_main", &Area::bgTileThemeMain)
    .def_readonly("bg_tile_theme_aux", &Area::bgTileThemeAux)
    .def_readonly("bg_pal_theme", &Area::bgPalTheme)
  ;

  //TEMP: How do we want to handle static arrays?
  def("sprites", &AreaSprByPhase, return_value_policy<reference_existing_object>());


  class_<std::deque<Area> >("AreaDeque")
    .def("size", &std::deque<Area>::size)

    //Enable iterators
    .def("__getitem__", &AreaDeque__get_item__, return_value_policy<reference_existing_object>())
    .def("__setitem__", &AreaDeque__set_item__)
  ;
}

