#include <cmath>
#include <boost/python.hpp>
#include <fstream>

#include "loaders.h"
#include "zeldagame.h"
#include "graphics.h"

using namespace boost::python;
using namespace Loaders;


//Stub functions which are located in other cpp files. Shares the loading effort.
void export_graphics();
void export_overworld();
void export_text();
void export_containers();


namespace {
//Helper method
ExpandInfo* MakeExpandInfo(const char* filename) {
  std::ifstream in(filename);
  return new ExpandInfo(in);
}

//Helper method
//I'm leaving this in for now, since std::dequeue access might be slow
Overworld::Map16Tile& GetMap16FromMap8(ZeldaGame& game, uint16_t map16Val) {
  return game.map16To8.at(map16Val);
}
} //End anon namespace


//Wrapper for boost::python access
BOOST_PYTHON_MODULE(blackmagic)
{
  //The Graphics namespace
  export_graphics();

  //The overworld namespace
  export_overworld();

  //The text namespace
  export_text();

  //The containers namespace
  export_containers();

  //Convenience functions. (Can remove/move later)
  def("MakeExpandInfo", &MakeExpandInfo, return_value_policy<manage_new_object>());
  def("GetMap16FromMap8", &GetMap16FromMap8, return_value_policy<reference_existing_object>());

  //"ExpandInfo" is used to handle loading from a file.
  class_<ExpandInfo>("ExpandInfo", init<std::ifstream&>())
  ;

  //The "ZeldaGame" class is filled with all relevant information upon loading.
  class_<ZeldaGame>("ZeldaGame")
    .def("Test2DCompression", &ZeldaGame::Test2DCompression)

    .def_readonly("messages", &ZeldaGame::messages)
    .def_readonly("areas", &ZeldaGame::areas)
  ;

  //The "RomLoader" class is used to expand a binary file into a "ZeldaGame" object
  class_<RomLoader>("RomLoader", init<ExpandInfo&>())
    .def("LoadGame", &RomLoader::LoadGame)
    .def("LoadDialogue", &RomLoader::LoadDialogue)
  ;




 

}


