#pragma once

#include <QMainWindow>
#include <QThread>
#include <QPushButton>
#include <vector>

namespace Ui {
class MainWindow;
}

//Helper class; uses RAII to block/unblock signals. All Qt signals are blocked
//  for the widget passed into the constructor while this class is live.
class BlockAllQtSignals {
public:
    BlockAllQtSignals(std::initializer_list<QWidget*> widgets) {
        for (auto* w : widgets) {
            if (w) {
                blocked.push_back(w);
                w->blockSignals(true);
            }
        }
    }
    ~BlockAllQtSignals() {
        for (auto w : blocked) {
            w->blockSignals(false);
        }
    }
private:
    BlockAllQtSignals() {}
    std::vector<QWidget*> blocked;
};


//This is getting silly.
struct bm_stuff;

//Helper thread for loading the ROM
class LoadRomThread : public QThread {
     Q_OBJECT
public:
     LoadRomThread(bm_stuff& stuff) : stuff(&stuff) {}
private:
     void run();
     bm_stuff* stuff;
};

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent=NULL);
    ~MainWindow();


private slots:
    void on_actionNew_triggered();
    void on_actionOpen_triggered();
    void romLoaded();
    void setGroupFromMetaGroup(int index);
    void setGroupFromGroup(int index);
    void setPalFromCount(int index);
    void setMapFromId(int index);

private:
    Ui::MainWindow *ui;
  //  std::vector<QPushButton*> colors; //Temp

    void openROM_Thread();
};

