#include <QtGui>
#include "ColorLister.hpp"

namespace {
    const int boxSize = 32; //TODO: Customizable, of course!
} //End anon namespace


ColorLister::ColorLister(QWidget *parent) : QFrame(parent),
    rowCutoff(8), minRows(4), minCols(8), defCellColor(QColor::fromRgb(0xFF, 0xFF, 0xFF))
{
    initUI();
}



void ColorLister::initUI()
{

}

QSize ColorLister::sizeHint () const
{
    //TODO: Use a more sophisticated method that takes the rowCutoff, etc. into account.
    int cols = minCols;
    int rows = minRows;
    return QSize(cols*(boxSize+2)+1, rows*(boxSize+2)+1);
}

void ColorLister::addColor(const QColor& clr)
{
    colors.push_back(clr);
    this->repaint();
}

void ColorLister::clearColors()
{
    colors.clear();
    this->repaint();
}


void ColorLister::paintEvent(QPaintEvent *e)
{
    //Dispatch to parent; get a painter
    QFrame::paintEvent(e);
    QPainter qp(this);

    //Calculate the number of rows/columns to show.
    //TODO: Use a more sophisticated method that takes the rowCutoff, etc. into account.
    int cols = minCols;
    int rows = minRows;
    int width = this->width();
    int height = this->height();

    //Fill the background
    qp.fillRect(0, 0, width, height, QColor::fromRgb(0xDDDDDD));

    //Prepare to draw.
    int xOffset = 0;
    int yOffset = 0;
    int numInRow = 0; //TODO: Later there are several stopping criteria
    qp.setPen(QPen(QColor::fromRgb(0x000000)));

    //Draw each box. For now this is simple.
    const int maxCell = std::max(minRows * minCols, static_cast<int>(colors.size()));
    for (int i=0; i<maxCell; i++) {
        QColor clr = i<colors.size() ? colors.at(i) : defCellColor;
        qp.fillRect(xOffset, yOffset, boxSize, boxSize, clr);
        qp.drawRect(xOffset, yOffset, boxSize, boxSize);
        xOffset += boxSize + 2;
        numInRow++;

        //Are we at a new row?
        if (numInRow >= rowCutoff) {
            xOffset = 0;
            yOffset += boxSize + 2;
            numInRow = 0;
        }
    }
}

















