#pragma once

#include <QWidget>
#include <QFrame>
#include <vector>

/**
 * Similar to the color picker, but geared towards simply showing off colors.
 */
class ColorLister : public QFrame {
    Q_OBJECT

public:
    ColorLister(QWidget *parent = 0);

    ///Add a new Color to this component. May trigger a resize and repaint.
    void addColor(const QColor& clr);

    ///Remove all known Colors from this component.
    void clearColors();

    ///Reimplemented to size properly
    QSize sizeHint () const;

protected:
    void paintEvent(QPaintEvent *e);

private:
    void initUI();

private:
    ///Array of colors (stored in one dimension)
    std::vector<QColor> colors;

    ///How many colors to show per row. This determines the component's width.
    ///If this is 0, the number of cells per row is determined the component's (maximum) width
    int rowCutoff;

    ///Minimum number of rows and columns to show when the color list is empty or there are an inadequate number of colors.
    //@{
    int minRows;
    int minCols;
    //@}

    ///Default background color for empty cells introduced by "minRows" and "minCols"
    QColor defCellColor;


};
