#include "FixedGridGraphicsView.hpp"

#include <QRectF>
#include <QRegion>
#include <QPainter>
#include <QPaintEvent>

#include <stdexcept>

#include "controls/FixedGridGraphicsScene.hpp"


FixedGridGraphicsView::FixedGridGraphicsView(QWidget* parent) : QGraphicsView(parent)
{
}


/*bool FixedGridGraphicsView::event (QEvent* event)
{
    return QGraphicsView::event(event);
}*/


void FixedGridGraphicsView::paintEvent (QPaintEvent* event)
{
    //If there's no scene, delegate this call to the parent (which will in turn
    //delegate it to the abstract scroll area
    if (!scene()) {
        QGraphicsView::paintEvent(event);
        return;
    }

    //Determine the exposed region in scene coordinates
    QRegion exposedRegion = event->region();
    QRectF exposedSceneRect = mapToScene(exposedRegion.boundingRect()).boundingRect();

    //Set up the painter around the current viewport.
    //QGraphicsView will request a viewport upon its creation, and will will not request a new one for its lifetime. The viewport is a QWidget.
    QPainter painter(viewport());

    //Save the current painter state.
    //(This is unnecessary (but harmless) if rubberbanding is disabled.
    painter.save();

    //Set up render hints
    //NOTE: By default we turn OFF text antialiasing.
    painter.setRenderHint(QPainter::TextAntialiasing, false);

    //Set up viewport transform
    if (isTransformed()) {
        painter.setWorldTransform(viewportTransform());
    }
    const QTransform& viewTransform = painter.worldTransform();

    //Now that the painter is set up, dispatch all drawing.
    performDrawing_Background(painter, exposedSceneRect);
    performDrawing_Items(painter, exposedRegion, exposedSceneRect, isTransformed(), viewTransform);
    performDrawing_Foreground(painter, exposedSceneRect);

    //Restore the painter state.
    // (This is unnecessary (but harmless) if rubberbanding is disabled)
    painter.restore();

    //Painting is DONE
    painter.end();
}


void FixedGridGraphicsView::performDrawing_Background(QPainter& painter, QRectF exposedSceneRect)
{
    if (cacheMode()==QGraphicsView::CacheNone) {
        //The simplest option: just repaint it. Assume painter state is preserved (may be slower, but is generally more stable)
        painter.save();
        drawBackground(&painter, exposedSceneRect);
        painter.restore();
    } else if (cacheMode()==QGraphicsView::CacheBackground) {
        throw std::runtime_error("Drawing with CacheBackground is currently disabled.");
    } else {
        throw std::runtime_error("Unexpected cacheMode");
    }
}


void FixedGridGraphicsView::performDrawing_Foreground(QPainter& painter, QRectF exposedSceneRect)
{
    drawForeground(&painter, exposedSceneRect);
}


FixedGridGraphicsScene *FixedGridGraphicsView::fg_scene()
{
    FixedGridGraphicsScene* fg_scene = dynamic_cast<FixedGridGraphicsScene*>(scene());
    if (!fg_scene) { throw std::runtime_error("FixedGridGraphicsView cannot use a generic Scene."); }
    return fg_scene;
}


void FixedGridGraphicsView::performDrawing_Items(QPainter& painter, QRegion exposedRegion, QRectF exposedSceneRect, bool viewTransformed, const QTransform& viewTransform)
{
    if (!(optimizationFlags()&QGraphicsView::IndirectPainting)) {
        //Call our internal drawing function
        int rectAdjust = (optimizationFlags() & QGraphicsView::DontAdjustForAntialiasing) ? 1 : 2;
        fg_scene()->drawItemsPrivate(&painter, rectAdjust, (viewTransformed ?&viewTransform : nullptr), &exposedRegion, viewport());

        //Ensure that the painter's state has been restored. This is simple, so just do it here;
        //no need for save() and restor() since only a few things changed
        painter.setOpacity(1.0);
        painter.setWorldTransform(viewTransform);
    } else {
        throw std::runtime_error("Drawing with IndirectPainting is currently disabled.");
    }
}







