#pragma once

#include "controls/FixedGridGraphicsView.hpp"

#include <map>
#include <vector>

#include "zeldagame.h"
#include "graphics.h"



class MapView : public FixedGridGraphicsView {
public:
    MapView(QWidget* parent=nullptr);

    //Basic functionality
    void newMap(int mapID);
    void renderAnnotation(Overworld::Map16Tile* tileQuad, int tileX, int tileY);
    void showHideMapLayer(bool show);
    void showHideAnnotLayer(bool show);

    //Might as well do something with those null button presses.
    void triCountAdd(int amount);

    //Subclassed functionality
    virtual void paintEvent (QPaintEvent* event);
    virtual void keyPressEvent (QKeyEvent* event);

    //Simple getters/setters
    void setGame(ZeldaGame* game) { this->game = game; }

    //These will inevitably be migrated to a separate class.
    static void BuildBgTiles(Graphics::BgTileMain bgMain, Graphics::BgTileAux bgAux, int count);
    static void BuildBgPalettes(Graphics::Group<7, 5> bg_prim, Graphics::Group<7, 3> bg_sec1, Graphics::Group<7, 3>* bg_sec2, Graphics::Group<7, 1> bg_tert);
private:
    //Convert a map8 tile into an 8x8 array of RGB pixel values
    static QImage* GetMap8Tile(ZeldaGame* game, Graphics::TileEntry* tileEntry, uint32_t bgColor);


    //Translate from input (x,y) values to offset values (raw values used by the underlying scene)
    std::pair<int, int> getTranslated(const std::pair<int, int>& point);

    void renderTile(Overworld::Map16Tile *tileQuad, int tileX, int tileY, uint32_t bgColor);
    void rebuildScene(int mapID);
    void clearSurrounding(int mapWidth, int mapHeight, bool ClearLeft, bool ClearRight, bool ClearTop, bool ClearBottom);
    std::pair<int, int> moveToEnd(const std::vector< std::vector<uint32_t> >& lookup, int startX, int startY, int xPlus, int yPlus);
    void loadSideMaps();

    //Returns a list of maps to use for looking up tiles.
    //Also includes the starting (tile) offset for the first map, in case a map spans multiple rows/columns
    //TODO: This doesn't work right for large maps (because they map to the LAST tile written)
    std::pair<std::vector<uint32_t>, int> getMapList(const std::vector< std::vector<uint32_t> >& lookup, int x, int y, int numTiles, bool horiz);

    //Using an offset and mapList populate a 'side' of the current viewing area.
    //If you run out of tiles, clear the remaining ones.
    //TODO: We can combine fillHoriz() and fillVert()... if we really want a headache, that is....
    void fillVertTiles(int srcY, const std::vector<uint32_t>& mapList, bool onRight);
    void fillHorizTiles(int srcX, const std::vector<uint32_t>& mapList, bool onBottom);

    ///If true, it expands the current map's visibility by 8x7 tiles on each side, loaded from the previous map.
    void showBorderAreas(bool show);

private:  //Static
    static std::pair<int, int> BorderSize;
    static QPixmap* QuestionTile;

    //Move these all out later.
    static std::map<std::string, QImage*> Map8TileCache;
    static std::vector<int> HalfPalSelect;
    static Graphics::TileSet<Graphics::Pixel3bpp, 0x200> BgTiles;
    static Graphics::Group<8, 16> BgPals;
    static Overworld::Area* CurrArea;
    static uint32_t CurrAreaID;
    static std::vector< std::vector<uint32_t> > AreaConnections; //2D array of area IDs
    static std::map<uint32_t, QPoint> AreaLookup; //<int> -> (x,y) in areaConnections

private:
    //Private properties
    ZeldaGame* game;
    bool mapLayerVis;
    bool annotLayerVis;
    bool validMap;
    bool hasLoadedSideMaps;
    std::pair<int, int> offset;

    //Purely for fun
    int triCount;
};



