
    #include <fstream>
    #include <sstream>
    #include <iomanip>
    #include <iostream>
    
    #include "containers.h"
    #include "loaders.h"
    
    #include "LzmaDec.h"
    #include "LzmaEnc.h"
    #include "Alloc.h"
    
// =============================================================================

    // extern from ExpandInfo.cpp
    int FileSize(std::ifstream &f);

    extern void DumpScript(ZeldaGame &zg);

    void CheckObjTypes(ZeldaGame &zg)
    {
        using namespace Dungeon;
        using namespace std;
        
        int roomNum = 0;
        
        ofstream o("./doors.log");
        
        // -----------------------------
        
        for
        (
            RoomDeque::iterator ia = zg.dungRooms.begin();
            ia != zg.dungRooms.end();
            ++ia)
        {
            Room &r = ( ia->GetRef() );
            
            // -----------------------------
            
            for(int layer = 0; layer < 3; ++layer)
            {
                DoorList &dList = r.objects[layer].doors;
                
                // -----------------------------
                
                for
                (
                    DoorList::iterator ib = dList.begin();
                    ib != dList.end();
                    ++ib)
                {
                    
                    o << hex << setfill('0') << uppercase << setw(0) << setprecision(8);
                    o << "Room: 0x" << setw(4);
                    o << roomNum << flush << ", ";
                    o << "Dir:  0x" << setw(4);
                    o << ib->GetDirection() << setw(0) << ", ";
                    o << "Pos:  0x" << setw(4);
                    o << static_cast<int>(ib->GetPosition()) << setw(0) << ", ";
                    o << "Typ:  0x" << setw(4);
                    o << static_cast<int>(ib->GetType()) << setw(0) << ", ";
                    o << "Ext:  0x" << setw(4);
                    o << static_cast<int>(ib->GetExtra()) << setw(0) << endl;
                }
            }
            
            roomNum += 1;
        }
    }
    
// =============================================================================
    
    void DumpAreas(ZeldaGame &zg)
    {
        using namespace std;
        using namespace Overworld;
        
        int byteCount = 0;
        
        Buffer b(0x02);
        
        ofstream obin("./areas.bin", ios::binary);
        ofstream otxt("./areas.txt");
        
        // --------------------------
        
        for(int i = 0; i < zg.areas.size(); ++i)
        {
            Area &a = zg.areas[i];
            
            std::stringstream s;
            
            s << "./areas" << i << ".bin";
            
            ofstream oeach(s.str(), ios::binary);
            
            // ------------------------
            
            otxt << byteCount << endl;
            
            for(int y = 0; y < a.areaMap.Height(); ++y)
            {
                for(int x = 0; x < a.areaMap.Width(); ++x)
                {
                    Map16Slot m16 = a.areaMap(x, y);
                    
                    if( m16.IsEmpty() == false )
                    {
                        b.Put2Bytes((uint32_t) 0,
                                    (Map16Val) m16);
                        
                        obin.write((const char*) b.contents,
                                   b.length);
                        
                        oeach.write((const char*) b.contents,
                                    b.length);
                        
                        byteCount += 2;
                    }
                }
            }
            
            oeach.close();
        }
        
        obin.close();
        otxt.close();
    }

// =============================================================================
    
    size_t CheckDungeonCompressRatio(ZeldaGame &zg)
    {
        using namespace Dungeon;
        
        size_t totalBytes = 0;
        
        for(int i = 0; i < 8; ++i)
        {
            totalBytes += zg.m_layouts[i].GetRef().size() * 3;
            
            // estimated size of terminators 0xfff0 and 0xffff
            totalBytes += 2;
        }
        
        for(int i = 0; i < zg.dungRooms.size(); ++i)
        {
            Room &r = ( zg.dungRooms[i].GetRef() );
            
            totalBytes += r.torches.size()   * 2;
            totalBytes += r.chests.size()    * 3;
            totalBytes += r.bigChests.size() * 3;
            totalBytes += r.blocks.size()    * 2;
            totalBytes += r.locks.size()     * 3;
            totalBytes += r.stairs.size()    * 3;
            totalBytes += r.stars.size()     * 3;
            
            for(int layer = 0; layer < 3; ++layer)
            {
                totalBytes += r.objects[layer].type1.size() * 3;
                totalBytes += r.objects[layer].doors.size() * 2;
                
                // estimated size of terminators 0xfff0 and 0xffff
                totalBytes += 4;
            }
        }
        
        return totalBytes;
    }

// =============================================================================

    static void *SzAlloc(void *p, size_t size) { p = p; return MyAlloc(size); }
    static void SzFree(void *p, void *address) { p = p; MyFree(address); }
    
// =============================================================================

    void TestLzmaEncode()
    {
        using std::ifstream;
        using std::ofstream;
        using std::ios;
        
        const Bool writeEndMark = True;
        
        Byte *dest         = nullptr;
        Byte *src          = nullptr;
        Byte *propsEncoded = nullptr;
        
        SizeT destLen   = 0x100000;
        SizeT srcLen    = 0;
        SizeT propsSize = LZMA_PROPS_SIZE;
        
        CLzmaEncProps props;
        
        ISzAlloc alloc = { SzAlloc, SzFree };
        
        ifstream f("./areas.bin", ios::binary);
        
        ofstream f2("./areas.lzma", ios::binary);
        
        // ---------------------------------
        
        srcLen = FileSize(f);
        
        src          = new Byte[srcLen];
        propsEncoded = new Byte[destLen];
        
        dest = (propsEncoded + LZMA_PROPS_SIZE);
        
        f.read( (char*) src, srcLen);
        
        // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        
        LzmaEncProps_Init(&props);
        
        props.dictSize = 0x2000;
        props.pb = 1;
        
        LzmaEncode(dest,
                   &destLen,
                   src,
                   srcLen,
                   &props,
                   propsEncoded,
                   &propsSize,
                   writeEndMark,
                   NULL,
                   &alloc,
                   &alloc);
        
        f2.write( (char*) propsEncoded, destLen + propsSize);
        
        f.close();
        f2.close();
        
        delete[] src;
        delete[] propsEncoded;
    }

// =============================================================================
        
    void TestLzmaDecode()
    {
        using std::ifstream;
        using std::ofstream;
        using std::ios;
        
        Byte *dest     = nullptr;
        Byte *propData = nullptr;
        Byte *src      = nullptr;
        
        unsigned propSize = 5;
        
        SizeT srcLen  = 0;
        SizeT destLen = 0x100000;
        
        ELzmaFinishMode finishMode;
        
        ELzmaStatus status;
        
        ISzAlloc alloc = { SzAlloc, SzFree };
        
        ifstream f("./areas.lzma", ios::binary);
        ofstream f2("./areas.delzma", ios::binary);
        
        // -----------------------------
        
        srcLen = FileSize(f);
        
        dest     = new Byte[destLen];
        propData = new Byte[srcLen];
        
        f.read( (char*) propData, srcLen);
        
        src = propData + propSize;
        
        srcLen -= propSize;
        
        LzmaDecode(dest,
                   &destLen,
                   (const Byte*) src,
                   &srcLen,
                   propData,
                   propSize,
                   LZMA_FINISH_ANY,
                   &status,
                   &alloc);
        
        f2.write( (char*) dest, destLen);
        
        f.close();
        f2.close();
        
        delete[] dest;
        delete[] propData;
        
    }
    
// =============================================================================
    
    int main(int argc, const char* argv[] )
    {
        using namespace std;
        using namespace Loaders;
        using namespace Dungeon;
        using namespace Overworld;
        using namespace Containers;
        
        ifstream f;
        
        ZeldaGame zg;
        
        // -----------------------------
        
        try
        {
            f.open(argv[1], ios::binary | ios::in);
        }
        catch(ios_base::failure failure)
        {
            // Failed due to no path, bad path, or whatever reason.
            return -1;
        }
        
        ExpandInfo e(f);
        
        if(e.CheckForExpansion())
        {
            ModernLoader m(zg, e);
            
            m.LoadGame();
        }
        else
        {
            RomLoader r(zg, e);
            
            r.LoadGame();
        }
        
    #if 1
        CheckDungeonCompressRatio(zg);
        
        DumpAreas(zg);
        DumpScript(zg);
        
        TestLzmaEncode();
        TestLzmaDecode();
        
        int numStairs = 0;
        
        for(int i = 0; i < zg.dungRooms.size(); ++i)
        {
            Room r = zg.dungRooms[i].Get();
            
            printf("\nRoom: %u", i);
            
            for(StairList::iterator si = r.stairs.begin(); si != r.stairs.end(); ++si)
            {
                Staircase &sc = (*si);
                
                numStairs += 1;
                
                printf("\n    x: %u, y: %u, floor: %s, target room: %u",
                       sc.m_pos.GetX(),
                       sc.m_pos.GetY(),
                       (sc.m_pos.GetFloor() == BG2) ? "BG2" : "BG1",
                       sc.m_targetRoom);
            }
            
            numStairs += r.stairs.size();
        }
        
        int numHoles = 0;
        int numEntr  = 0;
        int numExits = 0;
        
        for(AreaDeque::iterator it = zg.areas.begin(); it != zg.areas.end(); ++it)
        {
            numHoles += it->holes.size();
            numEntr  += it->entr.size();
            numExits += it->exits.size();
        }
        
        for(int i = 0; i < zg.areas.size(); ++i)
        {
            Area &a = zg.areas[i];
            
            Map16Buf &am = a.areaMap;
            
            for(uint8_t x = 0; x < am.Width(); ++x)
            {
                for(uint8_t y = 0; y < am.Height(); ++y)
                {
                    Show<uint16_t> deletemetoo;
                    Slot<uint16_t> deleteme;
                    
                    if( !deleteme.IsEmpty() )
                    {
                        deleteme.Clear();
                    }
                    
                    Map16Val v;
                    Map16Tile t;
                    Map16Slot &slot = am(x, y);
                    
                    if( slot.IsEmpty() )
                    {
                        continue;
                    }
                    
                    v = slot;
                    
                    if( v >= zg.map16To8.size() )
                    {
                        continue;
                    }
                    
                    t = zg.map16To8[v];
                    
                    for(int k = 0; k < 4; ++k)
                    {
                        TileEntry t8 = t[k];
                        
                        uint16_t chr = (t8.GetChr() & 0x03FF);
                        
                        if(chr >= 0x200)
                        {
                            printf("\nerror in room: %i, x: %i, y: %i", i, x, y);
                            
                            continue;
                        }
                        
                        uint8_t attr = e.m_rom->GetByte(0x71459 + chr);
                        
                        if(attr == 0x4E)
                        {
                            printf("\narea: %i, x: %i, y: %i", i, x, y);
                        }
                    }
                }
            }
        }
        
        CheckObjTypes(zg);
        
    #endif 
    
    #if 1
        int durp;
        
        std::cin >> durp;
    #endif
    }

// =============================================================================
